import urllib.parse

from benchling_api_client.v2.stable.client import Client


def replace_client_path(client: Client, base_path: str) -> Client:
    """Override a client's base URL with a new path. Does not update scheme, host, or other URL parts."""
    parsed = urllib.parse.urlparse(client.base_url)
    # _replace is not private, it's part of the NamedTuple API but prefixed _ to avoid conflicts
    updated_url = parsed._replace(path=base_path)
    return client.with_base_url(updated_url.geturl())


def v2_stable_client(client: Client) -> Client:
    """Override a client's base URL with a v2 stable path."""
    return replace_client_path(client, "api/v2")


def v2_alpha_client(client: Client) -> Client:
    """Override a client's base URL with a v2-alpha path."""
    return replace_client_path(client, "api/v2-alpha")


def v2_beta_client(client: Client) -> Client:
    """Override a client's base URL with a v2-beta path."""
    return replace_client_path(client, "api/v2-beta")



from datetime import datetime
from typing import Iterable, List, Optional, Union

from benchling_api_client.v2.stable.api.apps import (
    archive_app_canvases,
    archive_benchling_apps,
    bulk_create_app_configuration_items,
    bulk_update_app_configuration_items,
    create_app_canvas,
    create_app_configuration_item,
    create_app_session,
    create_benchling_app,
    get_app_canvas,
    get_app_configuration_item_by_id,
    get_app_session_by_id,
    get_benchling_app_by_id,
    list_app_canvases,
    list_app_configuration_items,
    list_app_sessions,
    list_benchling_apps,
    patch_benchling_app,
    unarchive_app_canvases,
    unarchive_benchling_apps,
    update_app_canvas,
    update_app_configuration_item,
    update_app_session,
)
from benchling_api_client.v2.types import Response

from benchling_sdk.errors import AppSessionClosedError, raise_for_status
from benchling_sdk.helpers.constants import _translate_to_string_enum
from benchling_sdk.helpers.decorators import api_method
from benchling_sdk.helpers.logging_helpers import log_deprecation
from benchling_sdk.helpers.pagination_helpers import NextToken, PageIterator
from benchling_sdk.helpers.response_helpers import model_from_detailed
from benchling_sdk.helpers.serialization_helpers import none_as_unset, optional_array_query_param
from benchling_sdk.models import (
    AppCanvas,
    AppCanvasCreate,
    AppCanvasesArchivalChange,
    AppCanvasesArchive,
    AppCanvasesArchiveReason,
    AppCanvasesPaginatedList,
    AppCanvasesUnarchive,
    AppCanvasUpdate,
    AppConfigItem,
    AppConfigItemBulkUpdate,
    AppConfigItemCreate,
    AppConfigItemsBulkCreateRequest,
    AppConfigItemsBulkUpdateRequest,
    AppConfigItemUpdate,
    AppConfigurationPaginatedList,
    AppSession,
    AppSessionCreate,
    AppSessionsPaginatedList,
    AppSessionUpdate,
    AsyncTaskLink,
    BenchlingApp,
    BenchlingAppCreate,
    BenchlingAppsArchivalChange,
    BenchlingAppsArchive,
    BenchlingAppsArchiveReason,
    BenchlingAppsPaginatedList,
    BenchlingAppsUnarchive,
    BenchlingAppUpdate,
    ListAppCanvasesEnabled,
    ListAppCanvasesSort,
    ListAppConfigurationItemsSort,
    ListAppSessionsSort,
    ListBenchlingAppsSort,
)
from benchling_sdk.services.v2.base_service import BaseService


class AppService(BaseService):
    """
    Apps.

    Create and manage Apps on your tenant.

    See https://benchling.com/api/reference#/Apps
    """

    @api_method
    def _apps_page(
        self,
        *,
        page_size: Optional[int] = 50,
        next_token: Optional[str] = None,
        sort: Optional[ListBenchlingAppsSort] = None,
        ids: Optional[Iterable[str]] = None,
        modified_at: Optional[str] = None,
        name: Optional[str] = None,
        name_includes: Optional[str] = None,
        namesany_of: Optional[Iterable[str]] = None,
        namesany_ofcase_sensitive: Optional[Iterable[str]] = None,
        creator_ids: Optional[Iterable[str]] = None,
        member_of: Optional[Iterable[str]] = None,
        admin_of: Optional[Iterable[str]] = None,
    ) -> Response[BenchlingAppsPaginatedList]:
        response = list_benchling_apps.sync_detailed(
            client=self.client,
            page_size=none_as_unset(page_size),
            next_token=none_as_unset(next_token),
            sort=none_as_unset(sort),
            ids=none_as_unset(optional_array_query_param(ids)),
            modified_at=none_as_unset(modified_at),
            name=none_as_unset(name),
            name_includes=none_as_unset(name_includes),
            namesany_of=none_as_unset(optional_array_query_param(namesany_of)),
            namesany_ofcase_sensitive=none_as_unset(optional_array_query_param(namesany_ofcase_sensitive)),
            creator_ids=none_as_unset(optional_array_query_param(creator_ids)),
            member_of=none_as_unset(optional_array_query_param(member_of)),
            admin_of=none_as_unset(optional_array_query_param(admin_of)),
        )
        raise_for_status(response)
        return response  # type: ignore

    def list_apps(
        self,
        *,
        page_size: Optional[int] = 50,
        sort: Optional[Union[str, ListBenchlingAppsSort]] = None,
        ids: Optional[Iterable[str]] = None,
        modified_at: Optional[str] = None,
        name: Optional[str] = None,
        name_includes: Optional[str] = None,
        namesany_of: Optional[Iterable[str]] = None,
        namesany_ofcase_sensitive: Optional[Iterable[str]] = None,
        creator_ids: Optional[str] = None,
        member_of: Optional[str] = None,
        admin_of: Optional[str] = None,
    ) -> PageIterator[BenchlingApp]:
        """
        List Apps.

        See https://benchling.com/api/reference#/Apps/listBenchlingApps
        """

        def api_call(next_token: NextToken) -> Response[BenchlingAppsPaginatedList]:
            return self._apps_page(
                page_size=page_size,
                sort=_translate_to_string_enum(ListBenchlingAppsSort, sort),
                ids=ids,
                modified_at=modified_at,
                name=name,
                name_includes=name_includes,
                namesany_of=namesany_of,
                namesany_ofcase_sensitive=namesany_ofcase_sensitive,
                creator_ids=creator_ids,
                member_of=member_of,
                admin_of=admin_of,
                next_token=next_token,
            )

        def results_extractor(body: BenchlingAppsPaginatedList) -> Optional[List[BenchlingApp]]:
            return body.apps

        return PageIterator(api_call, results_extractor)

    @api_method
    def get_by_id(self, app_id: str) -> BenchlingApp:
        """
        Get an App by ID.

        See https://benchling.com/api/reference#/Apps/getBenchlingAppByID
        """
        response = get_benchling_app_by_id.sync_detailed(client=self.client, app_id=app_id)
        return model_from_detailed(response)

    @api_method
    def create(self, app: BenchlingAppCreate) -> BenchlingApp:
        """
        Create an App (Deprecated).

        .. deprecated:: 1.8.0

        See https://benchling.com/api/reference#/Apps/createBenchlingApp
        """
        log_deprecation("apps.create")
        response = create_benchling_app.sync_detailed(client=self.client, json_body=app)
        return model_from_detailed(response)

    @api_method
    def update(self, app_id: str, app: BenchlingAppUpdate) -> BenchlingApp:
        """
        Update an App's metadata (Deprecated).

        .. deprecated:: 1.8.0

        See https://benchling.com/api/reference#/Apps/patchBenchlingApp
        """
        log_deprecation("apps.update")
        response = patch_benchling_app.sync_detailed(client=self.client, app_id=app_id, json_body=app)
        return model_from_detailed(response)

    @api_method
    def archive(
        self, app_ids: Iterable[str], reason: BenchlingAppsArchiveReason
    ) -> BenchlingAppsArchivalChange:
        """
        Archive Apps (Deprecated).

        .. deprecated:: 1.8.0

        See https://benchling.com/api/reference#/Apps/archiveBenchlingApps
        """
        log_deprecation("apps.archive")
        archive_request = BenchlingAppsArchive(app_ids=list(app_ids), reason=reason)
        response = archive_benchling_apps.sync_detailed(client=self.client, json_body=archive_request)
        return model_from_detailed(response)

    @api_method
    def unarchive(self, app_ids: Iterable[str]) -> BenchlingAppsArchivalChange:
        """
        Unarchive Apps (Deprecated).

        .. deprecated:: 1.8.0

        See https://benchling.com/api/reference#/Apps/unarchiveBenchlingApps
        """
        log_deprecation("apps.unarchive")
        unarchive_request = BenchlingAppsUnarchive(app_ids=list(app_ids))
        response = unarchive_benchling_apps.sync_detailed(client=self.client, json_body=unarchive_request)
        return model_from_detailed(response)

    @api_method
    def _list_app_configuration_items_page(
        self,
        *,
        app_id: Optional[str] = None,
        ids: Optional[Iterable[str]] = None,
        page_size: Optional[int] = 50,
        next_token: Optional[str] = None,
        modified_at: Optional[str] = None,
        sort: Optional[ListAppConfigurationItemsSort] = None,
    ) -> Response[AppConfigurationPaginatedList]:
        return list_app_configuration_items.sync_detailed(  # type: ignore
            client=self.client,
            app_id=none_as_unset(app_id),
            ids=none_as_unset(optional_array_query_param(ids)),
            page_size=none_as_unset(page_size),
            next_token=none_as_unset(next_token),
            modified_at=none_as_unset(modified_at),
            sort=none_as_unset(sort),
        )

    def list_app_configuration_items(
        self,
        *,
        app_id: Optional[str] = None,
        ids: Optional[Iterable[str]] = None,
        page_size: Optional[int] = 50,
        modified_at: Optional[str] = None,
        sort: Optional[Union[str, ListAppConfigurationItemsSort]] = None,
    ) -> PageIterator[AppConfigItem]:
        """
        Get app configuration items.

        See https://benchling.com/api/reference#/Apps/listAppConfigurationItems
        """

        def api_call(next_token: NextToken) -> Response[AppConfigurationPaginatedList]:
            return self._list_app_configuration_items_page(
                app_id=app_id,
                ids=ids,
                page_size=page_size,
                modified_at=modified_at,
                sort=_translate_to_string_enum(ListAppConfigurationItemsSort, sort),
                next_token=next_token,
            )

        def results_extractor(
            body: AppConfigurationPaginatedList,
        ) -> Optional[List[AppConfigItem]]:
            return body.app_configuration_items

        return PageIterator(api_call, results_extractor)

    @api_method
    def get_app_configuration_item_by_id(self, item_id: str) -> AppConfigItem:
        """
        Get app configuration.

        See https://benchling.com/api/reference#/Apps/getAppConfigurationItemById
        """
        response = get_app_configuration_item_by_id.sync_detailed(client=self.client, item_id=item_id)
        return model_from_detailed(response)

    @api_method
    def create_app_configuration_item(self, configuration_item: AppConfigItemCreate) -> AppConfigItem:
        """
        Create app configuration item.

        See https://benchling.com/api/reference#/Apps/createAppConfigurationItem
        """
        response = create_app_configuration_item.sync_detailed(
            client=self.client, json_body=configuration_item
        )
        return model_from_detailed(response)

    @api_method
    def update_app_configuration_item(
        self, item_id: str, configuration_item: AppConfigItemUpdate
    ) -> AppConfigItem:
        """
        Update app configuration item.

        See https://benchling.com/api/reference#/Apps/updateAppConfigurationItem
        """
        response = update_app_configuration_item.sync_detailed(
            client=self.client, item_id=item_id, json_body=configuration_item
        )
        return model_from_detailed(response)

    # TODO BNCH-52599 Add tests once platform supports archive
    # @api_method
    # def archive_app_configuration_items(
    #     self, item_ids: Iterable[str], reason: AppConfigItemsArchiveReason
    # ) -> AppConfigItemsArchivalChange:
    #     """
    #     Archive app configuration items.
    #
    #     See https://benchling.com/api/reference#/Apps/archiveAppConfigurationItems
    #     """
    #     archive_request = AppConfigItemsArchive(reason=reason, item_ids=list(item_ids))
    #     response = archive_app_configuration_items.sync_detailed(
    #         client=self.client, json_body=archive_request
    #     )
    #     return model_from_detailed(response)

    # TODO BNCH-52599 Add tests once platform supports archive
    # @api_method
    # def unarchive_app_configuration_items(self, item_ids: Iterable[str]) -> AppConfigItemsArchivalChange:
    #     """
    #     Unarchive app configuration items.
    #
    #     See https://benchling.com/api/reference#/Apps/unarchiveAppConfigurationItems
    #     """
    #     unarchive_request = AppConfigItemsUnarchive(item_ids=list(item_ids))
    #     response = unarchive_app_configuration_items.sync_detailed(
    #         client=self.client, json_body=unarchive_request
    #     )
    #     return model_from_detailed(response)

    @api_method
    def bulk_create_app_configuration_items(self, items: Iterable[AppConfigItemCreate]) -> AsyncTaskLink:
        """
        Bulk create app configuration items.

        See https://benchling.com/api/reference#/Apps/bulkCreateAppConfigurationItems
        """
        body = AppConfigItemsBulkCreateRequest(list(items))
        response = bulk_create_app_configuration_items.sync_detailed(client=self.client, json_body=body)
        return model_from_detailed(response)

    @api_method
    def bulk_update_app_configuration_items(self, items: Iterable[AppConfigItemBulkUpdate]) -> AsyncTaskLink:
        """
        Bulk update app configuration items.

        See https://benchling.com/api/reference#/Apps/bulkUpdateAppConfigurationItems
        """
        body = AppConfigItemsBulkUpdateRequest(list(items))
        response = bulk_update_app_configuration_items.sync_detailed(client=self.client, json_body=body)
        return model_from_detailed(response)

    # Canvases

    @api_method
    def create_canvas(self, canvas: AppCanvasCreate) -> AppCanvas:
        """
        Create an App Canvas that a Benchling App can write to and read user interaction from.

        See https://benchling.com/api/reference#/Apps/createAppCanvas
        """
        response = create_app_canvas.sync_detailed(
            client=self.client,
            json_body=canvas,
        )
        return model_from_detailed(response)

    @api_method
    def _canvases_page(
        self,
        app_id: Optional[str] = None,
        feature_id: Optional[str] = None,
        resource_id: Optional[str] = None,
        enabled: Optional[ListAppCanvasesEnabled] = None,
        archive_reason: Optional[str] = None,
        sort: Optional[Union[str, ListAppCanvasesSort]] = None,
        next_token: Optional[NextToken] = None,
        page_size: Optional[int] = None,
    ) -> Response[AppCanvasesPaginatedList]:
        response = list_app_canvases.sync_detailed(
            client=self.client,
            app_id=none_as_unset(app_id),
            feature_id=none_as_unset(feature_id),
            resource_id=none_as_unset(resource_id),
            enabled=none_as_unset(enabled),
            archive_reason=none_as_unset(archive_reason),
            sort=none_as_unset(_translate_to_string_enum(ListAppCanvasesSort, sort)),
            next_token=none_as_unset(next_token),
            page_size=none_as_unset(page_size),
        )
        raise_for_status(response)
        return response  # type: ignore

    # Removed until the API endpoints are released to GA (currently LA)

    # def list_canvases(
    #     self,
    #     app_id: Optional[str] = None,
    #     feature_id: Optional[str] = None,
    #     resource_id: Optional[str] = None,
    #     enabled: Optional[ListAppCanvasesEnabled] = None,
    #     archive_reason: Optional[str] = None,
    #     sort: Optional[Union[str, ListAppCanvasesSort]] = None,
    #     page_size: Optional[int] = None,
    # ) -> PageIterator[AppCanvas]:
    #     """
    #     List canvases.

    #     See https://benchling.com/api/reference#/Apps/listAppCanvases
    #     """

    #     def api_call(next_token: NextToken) -> Response[AppCanvasesPaginatedList]:
    #         return self._canvases_page(
    #             app_id=app_id,
    #             feature_id=feature_id,
    #             resource_id=resource_id,
    #             enabled=enabled,
    #             archive_reason=archive_reason,
    #             sort=sort,
    #             next_token=next_token,
    #             page_size=page_size,
    #         )

    #     def results_extractor(body: AppCanvasesPaginatedList) -> Optional[List[AppCanvas]]:
    #         return body.app_canvases

    #     return PageIterator(api_call, results_extractor)

    @api_method
    def get_canvas_by_id(self, canvas_id: str, returning: Optional[List[str]] = None) -> AppCanvas:
        """
        Get the current state of the App Canvas, including user input elements.

        See https://benchling.com/api/reference#/Apps/getAppCanvas
        """
        response = get_app_canvas.sync_detailed(
            client=self.client,
            canvas_id=canvas_id,
            returning=none_as_unset(optional_array_query_param(returning)),
        )
        return model_from_detailed(response)

    @api_method
    def update_canvas(self, canvas_id: str, canvas: AppCanvasUpdate) -> AppCanvas:
        """
        Update App Canvas.

        See https://benchling.com/api/reference#/Apps/updateAppCanvas
        """
        response = update_app_canvas.sync_detailed(
            client=self.client,
            canvas_id=canvas_id,
            json_body=canvas,
        )
        return model_from_detailed(response)

    @api_method
    def archive_canvases(
        self, canvas_ids: Iterable[str], reason: AppCanvasesArchiveReason
    ) -> AppCanvasesArchivalChange:
        """
        Archive App Canvases.

        See https://benchling.com/api/reference#/Apps/archiveAppCanvases
        """
        archive_request = AppCanvasesArchive(reason=reason, canvas_ids=list(canvas_ids))
        response = archive_app_canvases.sync_detailed(
            client=self.client,
            json_body=archive_request,
        )
        return model_from_detailed(response)

    @api_method
    def unarchive_canvases(self, canvas_ids: Iterable[str]) -> AppCanvasesArchivalChange:
        """
        Unarchive App Canvases.

        See https://benchling.com/api/reference#/Apps/unarchiveAppCanvases
        """
        unarchive_request = AppCanvasesUnarchive(canvas_ids=list(canvas_ids))
        response = unarchive_app_canvases.sync_detailed(client=self.client, json_body=unarchive_request)
        return model_from_detailed(response)

    # Sessions

    @api_method
    def create_session(self, session: AppSessionCreate) -> AppSession:
        """
        Create a new Benchling App session. Sessions cannot be archived once created.

        See https://benchling.com/api/v2/reference#/Apps/createAppSessions
        """
        response = create_app_session.sync_detailed(
            client=self.client,
            json_body=session,
        )
        return model_from_detailed(response)

    @api_method
    def get_session_by_id(self, session_id: str, returning: Optional[List[str]] = None) -> AppSession:
        """
        Get a Benchling App session.

        See https://benchling.com/api/v2/reference#/Apps/getAppSessionById
        """
        response = get_app_session_by_id.sync_detailed(
            client=self.client,
            id=session_id,
            returning=none_as_unset(optional_array_query_param(returning)),
        )
        return model_from_detailed(response)

    @api_method
    def update_session(self, session_id: str, session: AppSessionUpdate) -> AppSession:
        """
        Update Benchling App session.

        Raises AppSessionClosedError if trying to update a Session that has already been closed.

        See https://benchling.com/api/v2/reference#/Apps/updateAppSession
        """
        response = update_app_session.sync_detailed(
            client=self.client,
            id=session_id,
            json_body=session,
        )
        return model_from_detailed(response, error_types=[AppSessionClosedError])

    @api_method
    def _sessions_page(
        self,
        app_id: Optional[str] = None,
        created_at_lt: Optional[datetime] = None,
        created_at_gt: Optional[datetime] = None,
        created_at_lte: Optional[datetime] = None,
        created_at_gte: Optional[datetime] = None,
        modified_at_lt: Optional[datetime] = None,
        modified_at_gt: Optional[datetime] = None,
        modified_at_lte: Optional[datetime] = None,
        modified_at_gte: Optional[datetime] = None,
        sort: Optional[ListAppSessionsSort] = None,
        returning: Optional[Iterable[str]] = None,
        next_token: Optional[NextToken] = None,
        page_size: Optional[int] = None,
    ) -> Response[AppSessionsPaginatedList]:
        response = list_app_sessions.sync_detailed(
            client=self.client,
            app_id=none_as_unset(app_id),
            # TODO BNCH-70665 These should all be strings to be consistent with the rest of our API
            created_atlt=none_as_unset(created_at_lt),
            created_atgt=none_as_unset(created_at_gt),
            created_atlte=none_as_unset(created_at_lte),
            created_atgte=none_as_unset(created_at_gte),
            modified_atlt=none_as_unset(modified_at_lt),
            modified_atgt=none_as_unset(modified_at_gt),
            modified_atlte=none_as_unset(modified_at_lte),
            modified_atgte=none_as_unset(modified_at_gte),
            sort=none_as_unset(sort),
            returning=none_as_unset(optional_array_query_param(returning)),
            next_token=none_as_unset(next_token),
            page_size=none_as_unset(page_size),
        )
        raise_for_status(response)
        return response  # type: ignore

    # Removed until the API endpoints are released to GA (currently LA)

    # def list_sessions(
    #     self,
    #     app_id: Optional[str] = None,
    #     created_at_lt: Optional[datetime] = None,
    #     created_at_gt: Optional[datetime] = None,
    #     created_at_lte: Optional[datetime] = None,
    #     created_at_gte: Optional[datetime] = None,
    #     modified_at_lt: Optional[datetime] = None,
    #     modified_at_gt: Optional[datetime] = None,
    #     modified_at_lte: Optional[datetime] = None,
    #     modified_at_gte: Optional[datetime] = None,
    #     sort: Optional[ListAppSessionsSort] = None,
    #     returning: Optional[Iterable[str]] = None,
    #     page_size: Optional[int] = None,
    # ) -> PageIterator[AppSession]:
    #     """
    #     List all Benchling App sessions.

    #     See https://benchling.com/api/v2/reference#/Apps/listAppSessions
    #     """

    #     def api_call(next_token: NextToken) -> Response[AppSessionsPaginatedList]:
    #         return self._sessions_page(
    #             app_id=app_id,
    #             created_at_lt=created_at_lt,
    #             created_at_gt=created_at_gt,
    #             created_at_lte=created_at_lte,
    #             created_at_gte=created_at_gte,
    #             modified_at_lt=modified_at_lt,
    #             modified_at_gt=modified_at_gt,
    #             modified_at_lte=modified_at_lte,
    #             modified_at_gte=modified_at_gte,
    #             sort=sort,
    #             returning=returning,
    #             next_token=next_token,
    #             page_size=page_size,
    #         )

    #     def results_extractor(body: AppSessionsPaginatedList) -> Optional[List[AppSession]]:
    #         return body.app_sessions

    #     return PageIterator(api_call, results_extractor)



from typing import List, Optional, Union

from benchling_api_client.v2.stable.api.nucleotide_alignments import (
    create_consensus_nucleotide_alignment,
    create_template_nucleotide_alignment,
    delete_nucleotide_alignment,
    get_nucleotide_alignment,
    list_nucleotide_alignments,
)
from benchling_api_client.v2.types import Response

from benchling_sdk.errors import raise_for_status
from benchling_sdk.helpers.constants import _translate_to_string_enum
from benchling_sdk.helpers.decorators import api_method
from benchling_sdk.helpers.pagination_helpers import NextToken, PageIterator
from benchling_sdk.helpers.response_helpers import model_from_detailed
from benchling_sdk.helpers.serialization_helpers import none_as_unset, optional_array_query_param
from benchling_sdk.helpers.task_helpers import TaskHelper
from benchling_sdk.models import (
    ListNucleotideAlignmentsSort,
    NucleotideAlignment,
    NucleotideAlignmentsPaginatedList,
    NucleotideAlignmentSummary,
    NucleotideConsensusAlignmentCreate,
    NucleotideTemplateAlignmentCreate,
)
from benchling_sdk.services.v2.base_service import BaseService


class NucleotideAlignmentsService(BaseService):
    """
    Nucleotide Alignments.

    A Nucleotide Alignment is a Benchling object representing an alignment of multiple DNA and/or RNA sequences.

    See https://benchling.com/api/reference#/Nucleotide%20Alignments
    """

    @api_method
    def get_by_id(self, alignment_id: str) -> NucleotideAlignment:
        """
        Get a Nucleotide Alignment.

        See https://benchling.com/api/reference#/Nucleotide%20Alignments/getNucleotideAlignment
        """
        response = get_nucleotide_alignment.sync_detailed(client=self.client, alignment_id=alignment_id)
        return model_from_detailed(response)

    @api_method
    def _nucleotide_alignments_page(
        self,
        modified_at: Optional[str] = None,
        created_at: Optional[str] = None,
        name: Optional[str] = None,
        name_includes: Optional[str] = None,
        ids: Optional[List[str]] = None,
        names_any_of: Optional[List[str]] = None,
        names_any_of_case_sensitive: Optional[List[str]] = None,
        sequence_ids: Optional[List[str]] = None,
        sort: Optional[ListNucleotideAlignmentsSort] = None,
        page_size: Optional[int] = None,
        next_token: Optional[NextToken] = None,
    ) -> Response[NucleotideAlignmentsPaginatedList]:
        response = list_nucleotide_alignments.sync_detailed(
            client=self.client,
            modified_at=none_as_unset(modified_at),
            created_at=none_as_unset(created_at),
            name=none_as_unset(name),
            name_includes=none_as_unset(name_includes),
            ids=none_as_unset(optional_array_query_param(ids)),
            namesany_of=none_as_unset(optional_array_query_param(names_any_of)),
            namesany_ofcase_sensitive=none_as_unset(optional_array_query_param(names_any_of_case_sensitive)),
            sequence_ids=none_as_unset(optional_array_query_param(sequence_ids)),
            sort=none_as_unset(sort),
            page_size=none_as_unset(page_size),
            next_token=none_as_unset(next_token),
        )
        raise_for_status(response)
        return response  # type: ignore

    def list(
        self,
        modified_at: Optional[str] = None,
        created_at: Optional[str] = None,
        name: Optional[str] = None,
        name_includes: Optional[str] = None,
        ids: Optional[List[str]] = None,
        names_any_of: Optional[List[str]] = None,
        names_any_of_case_sensitive: Optional[List[str]] = None,
        sequence_ids: Optional[List[str]] = None,
        sort: Optional[Union[str, ListNucleotideAlignmentsSort]] = None,
        page_size: Optional[int] = None,
    ) -> PageIterator[NucleotideAlignmentSummary]:
        """
        List Nucleotide Alignments.

        See https://benchling.com/api/reference#/Nucleotide%20Alignments/listNucleotideAlignments
        """

        def api_call(next_token: NextToken) -> Response[NucleotideAlignmentsPaginatedList]:
            return self._nucleotide_alignments_page(
                modified_at=modified_at,
                created_at=created_at,
                name=name,
                name_includes=name_includes,
                ids=ids,
                names_any_of=names_any_of,
                names_any_of_case_sensitive=names_any_of_case_sensitive,
                sequence_ids=sequence_ids,
                sort=_translate_to_string_enum(ListNucleotideAlignmentsSort, sort),
                page_size=page_size,
                next_token=next_token,
            )

        def results_extractor(
            body: NucleotideAlignmentsPaginatedList,
        ) -> Optional[List[NucleotideAlignmentSummary]]:
            return body.alignments

        return PageIterator(api_call, results_extractor)

    @api_method
    def create_template_alignment(
        self, template_alignment: NucleotideTemplateAlignmentCreate
    ) -> TaskHelper[NucleotideAlignment]:
        """
        Create a template Nucleotide alignment.

        See https://benchling.com/api/reference#/Nucleotide%20Alignments/createTemplateNucleotideAlignment
        """
        response = create_template_nucleotide_alignment.sync_detailed(
            client=self.client, json_body=template_alignment
        )
        return self._task_helper_from_response(response, NucleotideAlignment)

    @api_method
    def create_consensus_alignment(
        self, consensus_alignment: NucleotideConsensusAlignmentCreate
    ) -> TaskHelper[NucleotideAlignment]:
        """
        Create a consensus Nucleotide alignment.

        See https://benchling.com/api/reference#/Nucleotide%20Alignments/createConsensusNucleotideAlignment
        """
        response = create_consensus_nucleotide_alignment.sync_detailed(
            client=self.client, json_body=consensus_alignment
        )
        return self._task_helper_from_response(response, NucleotideAlignment)

    @api_method
    def delete_alignment(self, alignment_id: str) -> None:
        """
        Delete a Nucleotide alignment.

        See https://benchling.com/api/reference#/Nucleotide%20Alignments/deleteNucleotideAlignment
        """
        response = delete_nucleotide_alignment.sync_detailed(client=self.client, alignment_id=alignment_id)
        raise_for_status(response)



from typing import Iterable, Optional

from benchling_api_client.v2.stable.api.entities import bulk_upsert_entities

from benchling_sdk.helpers.decorators import api_method
from benchling_sdk.helpers.response_helpers import model_from_detailed
from benchling_sdk.helpers.serialization_helpers import none_as_unset, optional_array_query_param
from benchling_sdk.models import AsyncTaskLink, EntitiesBulkUpsertRequest
from benchling_sdk.services.v2.base_service import BaseService


class EntityService(BaseService):
    """
    Entities.

    Entities include DNA and AA sequences, oligos, molecules, custom entities, and
    other biological objects in Benchling. Entities support schemas, tags, and aliases,
    and can be registered.

    See https://benchling.com/api/reference#/Entities
    """

    @api_method
    def bulk_upsert(
        self, bulk_upsert: EntitiesBulkUpsertRequest, returning: Optional[Iterable[str]] = None
    ) -> AsyncTaskLink:
        """
        Upsert many entities at once.

        All entities and their schemas must be within the same registry.

        This operation performs the following actions:

        1. Any existing objects are looked up in Benchling by the provided entity registry ID.
        2. Then, all objects are either created or updated accordingly, temporarily skipping any schema field links between objects.
        3. Schema field links between objects are populated according to the provided identifier. In the `value` field of the `Field` resource, the entity registry ID may be provided instead of the API ID if desired. You may link to objects being created in the same operation.
        4. Entities are registered, using the provided name and entity registry ID.

        If any action fails, the whole operation is canceled and no objects are created or updated.
        See https://benchling.com/api/reference#/Entities/bulkUpsertEntities
        """
        returning_string = optional_array_query_param(returning)
        response = bulk_upsert_entities.sync_detailed(
            client=self.client, json_body=bulk_upsert, returning=none_as_unset(returning_string)
        )
        return model_from_detailed(response)



from typing import Union

from benchling_sdk.models import (
    AaSequence,
    Blob,
    Box,
    Container,
    CustomEntity,
    DnaOligo,
    DnaSequence,
    Entry,
    Folder,
    Location,
    Mixture,
    Molecule,
    Plate,
    Request,
    RnaOligo,
    RnaSequence,
    WorkflowOutput,
    WorkflowTask,
)

# Taken from CHIP_SUPPORTED_COLUMN_TYPES in Benchling server
# Anything we miss, they can still embed the ID themselves in a message
ReferencedSessionLinkType = Union[
    AaSequence,
    Blob,
    Box,
    Container,
    CustomEntity,
    DnaSequence,
    DnaOligo,
    Entry,
    Folder,
    Location,
    Mixture,
    Plate,
    RnaOligo,
    Molecule,
    RnaSequence,
    Request,
    WorkflowOutput,
    WorkflowTask,
]



from benchling_api_client.v2.beta.api.apps import get_benchling_app_manifest, put_benchling_app_manifest
from benchling_api_client.v2.beta.models.benchling_app_manifest import BenchlingAppManifest

from benchling_sdk.helpers.decorators import api_method
from benchling_sdk.helpers.response_helpers import model_from_detailed
from benchling_sdk.services.v2.base_service import BaseService


class V2BetaAppService(BaseService):
    """
    V2-Beta Apps.

    Create and manage Apps on your tenant.

    https://benchling.com/api/v2-beta/reference#/Apps
    """

    @api_method
    def get_manifest(self, app_id: str) -> BenchlingAppManifest:
        """
        Get app manifest.

        See https://benchling.com/api/v2-beta/reference#/Apps/getBenchlingAppManifest
        """
        response = get_benchling_app_manifest.sync_detailed(client=self.client, app_id=app_id)
        return model_from_detailed(response)

    @api_method
    def update_manifest(self, app_id: str, manifest: BenchlingAppManifest) -> BenchlingAppManifest:
        """
        Update an app manifest.

        See https://benchling.com/api/v2-beta/reference#/Apps/putBenchlingAppManifest
        """
        response = put_benchling_app_manifest.sync_detailed(
            client=self.client, app_id=app_id, yaml_body=manifest
        )
        return model_from_detailed(response)



from typing import Iterable, Optional

from benchling_api_client.v2.stable.api.enzymes import list_enzymes
from benchling_api_client.v2.types import Response

from benchling_sdk.errors import raise_for_status
from benchling_sdk.helpers.decorators import api_method
from benchling_sdk.helpers.pagination_helpers import NextToken, PageIterator
from benchling_sdk.helpers.serialization_helpers import none_as_unset, optional_array_query_param
from benchling_sdk.models import Enzyme, EnzymesPaginatedList, ListEnzymesSort
from benchling_sdk.services.v2.base_service import BaseService


class EnzymeService(BaseService):
    """
    Enzymes.

    Restriction enzymes are curated by Benchling for operations such as Digests and Codon Optimization.

    See https://benchling.com/api/reference#/Enzymes
    """

    @api_method
    def _list_page(
        self,
        sort: Optional[ListEnzymesSort] = None,
        ids: Optional[Iterable[str]] = None,
        name: Optional[str] = None,
        name_includes: Optional[str] = None,
        names_any_of: Optional[Iterable[str]] = [],
        page_size: Optional[int] = None,
        next_token: Optional[NextToken] = None,
    ) -> Response[EnzymesPaginatedList]:
        response = list_enzymes.sync_detailed(
            client=self.client,
            ids=none_as_unset(optional_array_query_param(ids)),
            name=none_as_unset(name),
            name_includes=none_as_unset(name_includes),
            namesany_of=none_as_unset(optional_array_query_param(names_any_of)),
            next_token=none_as_unset(next_token),
            page_size=none_as_unset(page_size),
            sort=none_as_unset(sort),
        )
        return raise_for_status(response)

    def list(
        self,
        sort: Optional[ListEnzymesSort] = None,
        ids: Optional[Iterable[str]] = None,
        name: Optional[str] = None,
        name_includes: Optional[str] = None,
        names_any_of: Optional[Iterable[str]] = None,
        page_size: Optional[int] = None,
    ) -> PageIterator[Enzyme]:
        """
        List enzymes.

        See https://benchling.com/api/reference#/Enzymes/listEnzymes
        """

        def api_call(next_token: NextToken) -> Response[EnzymesPaginatedList]:
            return self._list_page(
                sort=sort,
                ids=ids,
                name=name,
                name_includes=name_includes,
                names_any_of=names_any_of,
                page_size=page_size,
                next_token=next_token,
            )

        return PageIterator(api_call, lambda result: result.enzymes)



from typing import Iterable, Optional

from benchling_api_client.v2.stable.api.monomers import (
    archive_monomers,
    create_monomer,
    list_monomers,
    unarchive_monomers,
    update_monomer,
)
from benchling_api_client.v2.types import Response

from benchling_sdk.errors import raise_for_status
from benchling_sdk.helpers.decorators import api_method
from benchling_sdk.helpers.pagination_helpers import NextToken, PageIterator
from benchling_sdk.helpers.response_helpers import model_from_detailed
from benchling_sdk.helpers.serialization_helpers import none_as_unset, optional_array_query_param
from benchling_sdk.models import (
    Monomer,
    MonomerCreate,
    MonomersArchivalChange,
    MonomersArchive,
    MonomersArchiveReason,
    MonomersPaginatedList,
    MonomersUnarchive,
    MonomerUpdate,
)
from benchling_sdk.services.v2.base_service import BaseService


class MonomerService(BaseService):
    """
    Monomers.

    Monomers are chemical building blocks with specified structures used to compose modified
    nucleotides. Note that monomer write endpoints require tenant admin permissions.

    See https://benchling.com/api/reference#/Monomers
    """

    @api_method
    def _list_page(
        self,
        page_size: Optional[int] = None,
        next_token: Optional[NextToken] = None,
        returning: Optional[Iterable[str]] = None,
    ) -> Response[MonomersPaginatedList]:
        response = list_monomers.sync_detailed(
            client=self.client,
            next_token=none_as_unset(next_token),
            page_size=none_as_unset(page_size),
            returning=none_as_unset(optional_array_query_param(returning)),
        )
        return raise_for_status(response)

    def list(
        self,
        page_size: Optional[int] = None,
        returning: Optional[Iterable[str]] = None,
    ) -> PageIterator[Monomer]:
        """
        List monomers.

        See https://benchling.com/api/reference#/Monomers/listMonomers
        """

        def api_call(next_token: NextToken) -> Response[MonomersPaginatedList]:
            return self._list_page(
                page_size=page_size,
                next_token=next_token,
                returning=returning,
            )

        return PageIterator(api_call, lambda result: result.monomers)

    @api_method
    def create(
        self,
        monomer: MonomerCreate,
        returning: Optional[Iterable[str]] = None,
    ) -> Monomer:
        """
        Create a monomer.

        See https://benchling.com/api/reference#/Monomers/createMonomer
        """
        response = create_monomer.sync_detailed(
            client=self.client,
            json_body=monomer,
            returning=none_as_unset(optional_array_query_param(returning)),
        )
        return model_from_detailed(response)

    @api_method
    def update(
        self,
        monomer_id: str,
        monomer: MonomerUpdate,
        returning: Optional[Iterable[str]] = None,
    ) -> Monomer:
        """
        Update a Monomer.

        See https://benchling.com/api/reference#/Monomers/updateMonomer
        """
        response = update_monomer.sync_detailed(
            client=self.client,
            monomer_id=monomer_id,
            json_body=monomer,
            returning=none_as_unset(optional_array_query_param(returning)),
        )
        return model_from_detailed(response)

    @api_method
    def archive(
        self,
        monomer_ids: Iterable[str],
        reason: MonomersArchiveReason,
    ) -> MonomersArchivalChange:
        """
        Archive Monomers.

        See https://benchling.com/api/reference#/Monomers/archiveMonomers
        """
        response = archive_monomers.sync_detailed(
            client=self.client,
            json_body=MonomersArchive(monomer_ids=list(monomer_ids), reason=reason),
        )
        return model_from_detailed(response)

    @api_method
    def unarchive(
        self,
        monomer_ids: Iterable[str],
    ) -> MonomersArchivalChange:
        """
        Unarchive Monomers.

        See https://benchling.com/api/reference#/Monomers/unarchiveMonomers
        """
        response = unarchive_monomers.sync_detailed(
            client=self.client,
            json_body=MonomersUnarchive(monomer_ids=list(monomer_ids)),
        )
        return model_from_detailed(response)



from benchling_api_client.v2.beta.api.app_definitions import (
    get_benchling_app_definition_version_manifest,
    put_benchling_app_definition_version_manifest,
)
from benchling_api_client.v2.beta.models.benchling_app_manifest import BenchlingAppManifest

from benchling_sdk.helpers.decorators import api_method
from benchling_sdk.helpers.response_helpers import model_from_detailed
from benchling_sdk.services.v2.base_service import BaseService


class V2BetaAppDefinitionService(BaseService):
    """
    V2-Beta App Definitions.

    Create and manage Benchling app definitions on your tenant.

    https://benchling.com/api/v2-beta/reference#/App%20Definitions
    """

    @api_method
    def get_version_manifest(self, app_def_id: str, version_id: str) -> BenchlingAppManifest:
        """
        Get manifest for an app definition version.

        See https://benchling.com/api/v2-beta/reference#/App%20Definitions/getBenchlingAppDefinitionVersionManifest
        """
        response = get_benchling_app_definition_version_manifest.sync_detailed(
            client=self.client, app_def_id=app_def_id, version_id=version_id
        )
        return model_from_detailed(response)

    @api_method
    def put_version_manifest(
        self, app_def_id: str, version_id: str, manifest: BenchlingAppManifest
    ) -> BenchlingAppManifest:
        """
        Create or update app definition version from a manifest.

        See https://benchling.com/api/v2-beta/reference#/App%20Definitions/putBenchlingAppDefinitionVersionManifest
        """
        response = put_benchling_app_definition_version_manifest.sync_detailed(
            client=self.client, app_def_id=app_def_id, version_id=version_id, yaml_body=manifest
        )
        return model_from_detailed(response)



from typing import Iterable, List, Optional

from benchling_api_client.v2.stable.api.datasets import (
    archive_datasets as api_client_archive_datasets,
    create_dataset,
    get_dataset,
    list_datasets,
    unarchive_datasets,
    update_dataset,
)
from benchling_api_client.v2.stable.models.dataset import Dataset
from benchling_api_client.v2.stable.models.dataset_create import DatasetCreate
from benchling_api_client.v2.stable.models.dataset_update import DatasetUpdate
from benchling_api_client.v2.stable.models.datasets_archival_change import DatasetsArchivalChange
from benchling_api_client.v2.stable.models.datasets_archive import DatasetsArchive
from benchling_api_client.v2.stable.models.datasets_archive_reason import DatasetsArchiveReason
from benchling_api_client.v2.stable.models.datasets_paginated_list import DatasetsPaginatedList
from benchling_api_client.v2.stable.models.datasets_unarchive import DatasetsUnarchive
from benchling_api_client.v2.types import Response

from benchling_sdk.errors import raise_for_status
from benchling_sdk.helpers.decorators import api_method
from benchling_sdk.helpers.pagination_helpers import NextToken, PageIterator
from benchling_sdk.helpers.response_helpers import model_from_detailed
from benchling_sdk.helpers.serialization_helpers import none_as_unset
from benchling_sdk.models import ListDatasetsSort
from benchling_sdk.services.v2.base_service import BaseService


class DatasetService(BaseService):
    """
    Datasets.

    Similar to Data frames, datasets in Benchling represent tabular data that is not schematized. Datasets are
    saved to folders within Benchling with additional metadata, making them accessible and searchable within
    Benchling. Each dataset actually contains a data frame, and a data frame is required to create a dataset.

    See https://benchling.com/api/v2/reference#/Datasets
    """

    @api_method
    def get_by_id(self, dataset_id: str) -> Dataset:
        """
        Get a dataset.

        See https://benchling.com/api/v2/reference#/Datasets/getDataset
        """
        response = get_dataset.sync_detailed(client=self.client, dataset_id=dataset_id)
        return model_from_detailed(response)

    @api_method
    def archive_datasets(
        self, dataset_ids: Iterable[str], reason: DatasetsArchiveReason
    ) -> DatasetsArchivalChange:
        """
        Archive Datasets.

        See https://benchling.com/api/reference#/Datasets/archiveDatasets
        """
        archive_request = DatasetsArchive(reason=reason, dataset_ids=list(dataset_ids))
        response = api_client_archive_datasets.sync_detailed(
            client=self.client,
            json_body=archive_request,
        )
        return model_from_detailed(response)

    @api_method
    def create(self, dataset: DatasetCreate) -> Dataset:
        """
        Create a dataset.

        See https://benchling.com/api/v2/reference#/Datasets/createDataset
        """
        response = create_dataset.sync_detailed(client=self.client, json_body=dataset)
        return model_from_detailed(response)

    @api_method
    def _datasets_page(
        self,
        page_size: Optional[int] = 50,
        next_token: Optional[str] = None,
        sort: Optional[ListDatasetsSort] = ListDatasetsSort.MODIFIEDAT,
        archive_reason: Optional[str] = None,
        created_at: Optional[str] = None,
        creator_ids: Optional[str] = None,
        folder_id: Optional[str] = None,
        mentioned_in: Optional[str] = None,
        modified_at: Optional[str] = None,
        name: Optional[str] = None,
        name_includes: Optional[str] = None,
        namesany_ofcase_sensitive: Optional[str] = None,
        namesany_of: Optional[str] = None,
        origin_ids: Optional[str] = None,
        ids: Optional[str] = None,
        display_ids: Optional[str] = None,
        returning: Optional[str] = None,
    ) -> Response[DatasetsPaginatedList]:
        response = list_datasets.sync_detailed(
            client=self.client,
            page_size=none_as_unset(page_size),
            next_token=none_as_unset(next_token),
            sort=none_as_unset(sort),
            archive_reason=none_as_unset(archive_reason),
            created_at=none_as_unset(created_at),
            creator_ids=none_as_unset(creator_ids),
            folder_id=none_as_unset(folder_id),
            mentioned_in=none_as_unset(mentioned_in),
            modified_at=none_as_unset(modified_at),
            name=none_as_unset(name),
            name_includes=none_as_unset(name_includes),
            namesany_ofcase_sensitive=none_as_unset(namesany_ofcase_sensitive),
            namesany_of=none_as_unset(namesany_of),
            origin_ids=none_as_unset(origin_ids),
            ids=none_as_unset(ids),
            display_ids=none_as_unset(display_ids),
            returning=none_as_unset(returning),
        )
        raise_for_status(response)
        return response  # type: ignore

    def list(
        self,
        *,
        page_size: Optional[int] = 50,
        sort: Optional[ListDatasetsSort] = ListDatasetsSort.MODIFIEDAT,
        archive_reason: Optional[str] = None,
        created_at: Optional[str] = None,
        creator_ids: Optional[str] = None,
        folder_id: Optional[str] = None,
        mentioned_in: Optional[str] = None,
        modified_at: Optional[str] = None,
        name: Optional[str] = None,
        name_includes: Optional[str] = None,
        namesany_ofcase_sensitive: Optional[str] = None,
        namesany_of: Optional[str] = None,
        origin_ids: Optional[str] = None,
        ids: Optional[str] = None,
        display_ids: Optional[str] = None,
        returning: Optional[str] = None,
    ) -> PageIterator[Dataset]:
        """
        List Datasets.

        See https://benchling.com/api/v2/reference#/Datasets/listDatasets
        """

        def api_call(next_token: NextToken) -> Response[DatasetsPaginatedList]:
            return self._datasets_page(
                page_size=page_size,
                next_token=next_token,
                sort=sort,
                archive_reason=archive_reason,
                created_at=created_at,
                creator_ids=creator_ids,
                folder_id=folder_id,
                mentioned_in=mentioned_in,
                modified_at=modified_at,
                name=name,
                name_includes=name_includes,
                namesany_ofcase_sensitive=namesany_ofcase_sensitive,
                namesany_of=namesany_of,
                origin_ids=origin_ids,
                ids=ids,
                display_ids=display_ids,
                returning=returning,
            )

        def results_extractor(body: DatasetsPaginatedList) -> Optional[List[Dataset]]:
            return body.datasets

        return PageIterator(api_call, results_extractor)

    @api_method
    def unarchive(self, dataset_ids: Iterable[str]) -> DatasetsArchivalChange:
        """
        Unarchive one or more Datasets.

        See https://benchling.com/api/reference#/Datasets/unarchiveDatasets
        """
        unarchive_request = DatasetsUnarchive(dataset_ids=list(dataset_ids))
        response = unarchive_datasets.sync_detailed(client=self.client, json_body=unarchive_request)
        return model_from_detailed(response)

    @api_method
    def update(self, dataset_id: str, dataset: DatasetUpdate) -> Dataset:
        """
        Update a Dataset.

        See https://benchling.com/api/reference#/Datasets/updateDataset
        """
        response = update_dataset.sync_detailed(client=self.client, dataset_id=dataset_id, json_body=dataset)
        return model_from_detailed(response)



from typing import List

from benchling_api_client.v2.stable.api.connect import (
    convert_to_asm as api_client_convert_to_asm,
    convert_to_csv as api_client_convert_to_csv,
    list_allotropy_vendors as api_client_list_allotropy_vendors,
)
from benchling_api_client.v2.stable.models.convert_to_asm import ConvertToASM
from benchling_api_client.v2.stable.models.convert_to_asm_response_200 import ConvertToASMResponse_200
from benchling_api_client.v2.stable.models.convert_to_csv import ConvertToCSV
from benchling_api_client.v2.stable.models.convert_to_csv_response_200_item import (
    ConvertToCSVResponse_200Item,
)

from benchling_sdk.errors import raise_for_status
from benchling_sdk.helpers.decorators import api_method
from benchling_sdk.helpers.response_helpers import model_from_detailed
from benchling_sdk.services.v2.base_service import BaseService


class ConnectService(BaseService):
    """
    Connect.

    Connect endpoints support Benchling Connect actions, like instrument data conversion.

    See https://benchling.com/api/reference#/Connect
    """

    @api_method
    def convert_to_asm(self, file_info: ConvertToASM) -> ConvertToASMResponse_200:
        """
        Convert an input blob or file containing instrument data to ASM (Allotrope Simple Model) JSON.

        May provide the name of the instrument vendor (see /connect/list-allotropy-vendors) or the ID of a
        connection associated with an instrument vendor.

        See https://benchling.com/api/reference#/Connect/convertToASM
        """
        response = api_client_convert_to_asm.sync_detailed(client=self.client, json_body=file_info)
        response = raise_for_status(response)
        return model_from_detailed(response)

    @api_method
    def convert_to_csv(self, file_info: ConvertToCSV) -> List[ConvertToCSVResponse_200Item]:
        """
        Convert a blob or file containing ASM, JSON, or instrument data to CSV.

        If the file is ASM JSON, specify either no transform type (in which case all transform types will be
        returned), a matching transform type for the ASM schema, or a custom JSON mapper config.

        If the file non-ASM JSON, must provide a JSON mapper config argument, which specifies how to map the
        JSON to CSV. Reach out to Benchling Support for more information about how to create a JSON mapper
        config.

        If the file is an instrument file, must also specify an instrument vendor. The file will be converted
        first to ASM JSON and then to CSV. Only the CSV output will be returned.

        May provide an AutomationOutputFile with CSV transform arguments configured to read the transform type
        or mapper config from.

        May provide a connection ID associated with an instrument to read the vendor from.

        See https://benchling.com/api/reference#/Connect/convertToCSV
        """
        response = api_client_convert_to_csv.sync_detailed(client=self.client, json_body=file_info)
        response = raise_for_status(response)
        return model_from_detailed(response)

    @api_method
    def list_allotropy_vendors(self) -> List[None]:
        """
        Return the list of available allotropy instrument vendor types.

        See https://benchling.com/api/reference#/Connect/listAllotropyVendors
        """
        response = api_client_list_allotropy_vendors.sync_detailed(client=self.client)
        response = raise_for_status(response)
        return model_from_detailed(response)



import datetime
from typing import Any, Dict, Iterable, List, Optional, Union

from benchling_api_client.v2.extensions import UnknownType
from benchling_api_client.v2.stable.api.workflow_tasks import (
    archive_workflow_tasks,
    bulk_copy_workflow_tasks,
    bulk_create_workflow_tasks,
    bulk_update_workflow_tasks,
    copy_workflow_task,
    create_workflow_task,
    get_workflow_task,
    list_workflow_tasks,
    unarchive_workflow_tasks,
    update_workflow_task,
)
from benchling_api_client.v2.types import Response

from benchling_sdk.errors import raise_for_status
from benchling_sdk.helpers.decorators import api_method
from benchling_sdk.helpers.pagination_helpers import NextToken, PageIterator
from benchling_sdk.helpers.response_helpers import model_from_detailed
from benchling_sdk.helpers.serialization_helpers import (
    none_as_unset,
    optional_array_query_param,
    schema_fields_query_param,
)
from benchling_sdk.models import (
    AsyncTaskLink,
    ListWorkflowTasksScheduledOn,
    WorkflowTask,
    WorkflowTaskArchiveReason,
    WorkflowTaskBulkCreate,
    WorkflowTaskBulkUpdate,
    WorkflowTaskCreate,
    WorkflowTasksArchivalChange,
    WorkflowTasksArchive,
    WorkflowTasksBulkCopyRequest,
    WorkflowTasksBulkCreateRequest,
    WorkflowTasksBulkUpdateRequest,
    WorkflowTasksPaginatedList,
    WorkflowTasksUnarchive,
    WorkflowTaskUpdate,
)
from benchling_sdk.services.v2.base_service import BaseService


class WorkflowTaskService(BaseService):
    """
    Workflow Tasks.

    Workflow tasks encapsulate a single unit of work.

    See https://benchling.com/api/reference#/Workflow%20Tasks
    """

    @api_method
    def get_by_id(self, workflow_task_id: str) -> WorkflowTask:
        """
        Get a workflow task.

        See https://benchling.com/api/reference#/Workflow%20Tasks/getWorkflowTask
        """
        response = get_workflow_task.sync_detailed(client=self.client, workflow_task_id=workflow_task_id)
        return model_from_detailed(response)

    @api_method
    def _workflow_tasks_page(
        self,
        ids: Optional[Iterable[str]] = None,
        workflow_task_group_ids: Optional[Iterable[str]] = None,
        schema_id: Optional[str] = None,
        status_ids: Optional[Iterable[str]] = None,
        assignee_ids: Optional[Iterable[str]] = None,
        watcher_ids: Optional[Iterable[str]] = None,
        responsible_team_ids: Optional[Iterable[str]] = None,
        execution_origin_ids: Optional[Iterable[str]] = None,
        execution_types: Optional[Iterable[str]] = None,
        schema_fields: Optional[Dict[str, Any]] = None,
        name: Optional[str] = None,
        name_includes: Optional[str] = None,
        creator_ids: Optional[Iterable[str]] = None,
        scheduled_on: Union[None, ListWorkflowTasksScheduledOn, datetime.date, UnknownType] = None,
        scheduled_on_lt: Optional[datetime.date] = None,
        scheduled_on_lte: Optional[datetime.date] = None,
        scheduled_on_gte: Optional[datetime.date] = None,
        scheduled_on_gt: Optional[datetime.date] = None,
        modified_at: Optional[str] = None,
        display_ids: Optional[Iterable[str]] = None,
        linked_item_ids_any_of: Optional[Iterable[str]] = None,
        linked_item_ids_all_of: Optional[Iterable[str]] = None,
        linked_item_ids_none_of: Optional[Iterable[str]] = None,
        archive_reason: Optional[str] = None,
        page_size: Optional[int] = None,
        next_token: Optional[NextToken] = None,
    ) -> Response[WorkflowTasksPaginatedList]:
        response = list_workflow_tasks.sync_detailed(
            client=self.client,
            ids=none_as_unset(optional_array_query_param(ids)),
            workflow_task_group_ids=none_as_unset(optional_array_query_param(workflow_task_group_ids)),
            schema_id=none_as_unset(schema_id),
            status_ids=none_as_unset(optional_array_query_param(status_ids)),
            assignee_ids=none_as_unset(optional_array_query_param(assignee_ids)),
            watcher_ids=none_as_unset(optional_array_query_param(watcher_ids)),
            responsible_team_ids=none_as_unset(optional_array_query_param(responsible_team_ids)),
            execution_origin_ids=none_as_unset(optional_array_query_param(execution_origin_ids)),
            execution_types=none_as_unset(optional_array_query_param(execution_types)),
            schema_fields=none_as_unset(schema_fields_query_param(schema_fields)),
            name=none_as_unset(name),
            name_includes=none_as_unset(name_includes),
            creator_ids=none_as_unset(optional_array_query_param(creator_ids)),
            scheduled_on=none_as_unset(scheduled_on),
            scheduled_onlt=none_as_unset(scheduled_on_lt),
            scheduled_onlte=none_as_unset(scheduled_on_lte),
            scheduled_ongte=none_as_unset(scheduled_on_gte),
            scheduled_ongt=none_as_unset(scheduled_on_gt),
            modified_at=none_as_unset(modified_at),
            display_ids=none_as_unset(optional_array_query_param(display_ids)),
            linked_item_idsany_of=none_as_unset(optional_array_query_param(linked_item_ids_any_of)),
            linked_item_idsall_of=none_as_unset(optional_array_query_param(linked_item_ids_all_of)),
            linked_item_idsnone_of=none_as_unset(optional_array_query_param(linked_item_ids_none_of)),
            archive_reason=none_as_unset(archive_reason),
            page_size=none_as_unset(page_size),
            next_token=none_as_unset(next_token),
        )
        raise_for_status(response)
        return response  # type: ignore

    def list(
        self,
        ids: Optional[Iterable[str]] = None,
        workflow_task_group_ids: Optional[Iterable[str]] = None,
        schema_id: Optional[str] = None,
        status_ids: Optional[Iterable[str]] = None,
        assignee_ids: Optional[Iterable[str]] = None,
        watcher_ids: Optional[Iterable[str]] = None,
        responsible_team_ids: Optional[Iterable[str]] = None,
        execution_origin_ids: Optional[Iterable[str]] = None,
        execution_types: Optional[Iterable[str]] = None,
        schema_fields: Optional[Dict[str, Any]] = None,
        name: Optional[str] = None,
        name_includes: Optional[str] = None,
        creator_ids: Optional[Iterable[str]] = None,
        scheduled_on: Union[None, ListWorkflowTasksScheduledOn, datetime.date, UnknownType] = None,
        scheduled_on_lt: Optional[datetime.date] = None,
        scheduled_on_lte: Optional[datetime.date] = None,
        scheduled_on_gte: Optional[datetime.date] = None,
        scheduled_on_gt: Optional[datetime.date] = None,
        modified_at: Optional[str] = None,
        display_ids: Optional[Iterable[str]] = None,
        linked_item_ids_any_of: Optional[Iterable[str]] = None,
        linked_item_ids_all_of: Optional[Iterable[str]] = None,
        linked_item_ids_none_of: Optional[Iterable[str]] = None,
        archive_reason: Optional[str] = None,
        page_size: Optional[int] = None,
    ) -> PageIterator[WorkflowTask]:
        """
        List workflow tasks.

        See https://benchling.com/api/reference#/Workflow%20Tasks/listWorkflowTasks
        """

        def api_call(next_token: NextToken) -> Response[WorkflowTasksPaginatedList]:
            return self._workflow_tasks_page(
                ids=ids,
                workflow_task_group_ids=workflow_task_group_ids,
                schema_id=schema_id,
                status_ids=status_ids,
                assignee_ids=assignee_ids,
                watcher_ids=watcher_ids,
                responsible_team_ids=responsible_team_ids,
                execution_origin_ids=execution_origin_ids,
                execution_types=execution_types,
                schema_fields=schema_fields,
                name=name,
                name_includes=name_includes,
                creator_ids=creator_ids,
                scheduled_on=scheduled_on,
                scheduled_on_lt=scheduled_on_lt,
                scheduled_on_lte=scheduled_on_lte,
                scheduled_on_gte=scheduled_on_gte,
                scheduled_on_gt=scheduled_on_gt,
                modified_at=modified_at,
                display_ids=display_ids,
                linked_item_ids_any_of=linked_item_ids_any_of,
                linked_item_ids_all_of=linked_item_ids_all_of,
                linked_item_ids_none_of=linked_item_ids_none_of,
                archive_reason=archive_reason,
                page_size=page_size,
                next_token=next_token,
            )

        def results_extractor(body: WorkflowTasksPaginatedList) -> Optional[List[WorkflowTask]]:
            return body.workflow_tasks

        return PageIterator(api_call, results_extractor)

    @api_method
    def create(self, workflow_task: WorkflowTaskCreate) -> WorkflowTask:
        """
        Create a new workflow task.

        See https://benchling.com/api/reference#/Workflow%20Tasks/createWorkflowTask
        """
        response = create_workflow_task.sync_detailed(client=self.client, json_body=workflow_task)
        return model_from_detailed(response)

    @api_method
    def update(self, workflow_task_id: str, workflow_task: WorkflowTaskUpdate) -> WorkflowTask:
        """
        Update a workflow task.

        See https://benchling.com/api/reference#/Workflow%20Tasks/updateWorkflowTask
        """
        response = update_workflow_task.sync_detailed(
            client=self.client, workflow_task_id=workflow_task_id, json_body=workflow_task
        )
        return model_from_detailed(response)

    @api_method
    def copy(self, workflow_task_id: str) -> WorkflowTask:
        """
        Copy workflow task.

        Creates a new workflow task with the same fields and assignee as the provided task and creates
        a relationship between the two tasks.

        See https://benchling.com/api/reference#/Workflow%20Tasks/copyWorkflowTask
        """
        response = copy_workflow_task.sync_detailed(
            client=self.client,
            workflow_task_id=workflow_task_id,
        )
        return model_from_detailed(response)

    @api_method
    def archive(
        self, workflow_task_ids: Iterable[str], reason: WorkflowTaskArchiveReason
    ) -> WorkflowTasksArchivalChange:
        """
        Archive one or more workflow tasks.

        See https://benchling.com/api/reference#/Workflow%20Tasks/archiveWorkflowTasks
        """
        archive_request = WorkflowTasksArchive(reason=reason, workflow_task_ids=list(workflow_task_ids))
        response = archive_workflow_tasks.sync_detailed(client=self.client, json_body=archive_request)
        return model_from_detailed(response)

    @api_method
    def unarchive(self, workflow_task_ids: Iterable[str]) -> WorkflowTasksArchivalChange:
        """
        Unarchive one or more workflow tasks.

        See https://benchling.com/api/reference#/Workflow%20Tasks/unarchiveWorkflowTasks
        """
        unarchive_request = WorkflowTasksUnarchive(workflow_task_ids=list(workflow_task_ids))
        response = unarchive_workflow_tasks.sync_detailed(client=self.client, json_body=unarchive_request)
        return model_from_detailed(response)

    @api_method
    def bulk_create(self, workflow_tasks: Iterable[WorkflowTaskBulkCreate]) -> AsyncTaskLink:
        """
        Create one or more workflow tasks.

        See https://benchling.com/api/reference#/Workflow%20Tasks/bulkCreateWorkflowTasks
        """
        body = WorkflowTasksBulkCreateRequest(list(workflow_tasks))
        response = bulk_create_workflow_tasks.sync_detailed(client=self.client, json_body=body)
        return model_from_detailed(response)

    @api_method
    def bulk_update(self, workflow_tasks: Iterable[WorkflowTaskBulkUpdate]) -> AsyncTaskLink:
        """
        Update one or more workflow tasks.

        See https://benchling.com/api/reference#/Workflow%20Tasks/bulkUpdateWorkflowTasks
        """
        body = WorkflowTasksBulkUpdateRequest(list(workflow_tasks))
        response = bulk_update_workflow_tasks.sync_detailed(client=self.client, json_body=body)
        return model_from_detailed(response)

    @api_method
    def bulk_copy(self, workflow_task_ids: Iterable[str]) -> AsyncTaskLink:
        """
        Bulk copy workflow tasks.

        Bulk creates new workflow tasks where each new task has the same fields and assignee as one of
        the provided tasks and creates a relationship between the provided task and its copy

        See https://benchling.com/api/reference#/Workflow%20Tasks/bulkCopyWorkflowTasks
        """
        body = WorkflowTasksBulkCopyRequest(list(workflow_task_ids))
        response = bulk_copy_workflow_tasks.sync_detailed(client=self.client, json_body=body)
        return model_from_detailed(response)



from __future__ import annotations

from typing import Dict, List, Optional, OrderedDict, Protocol, Tuple, Union

from benchling_api_client.v2.extensions import UnknownType
from ordered_set import OrderedSet

from benchling_sdk.apps.config.errors import (
    ConfigItemLinkedResourceError,
    InaccessibleConfigItemError,
    MissingRequiredConfigItemError,
    UnsupportedConfigItemError,
)
from benchling_sdk.apps.config.types import ConfigItemPath, ConfigurationReference
from benchling_sdk.benchling import Benchling
from benchling_sdk.models import (
    AppConfigItem,
    EntitySchemaAppConfigItem,
    FieldAppConfigItem,
    GenericApiIdentifiedAppConfigItem,
    InaccessibleResource,
    LinkedAppConfigResourceSummary,
    ListAppConfigurationItemsSort,
)


class _DefaultOrderedDict(OrderedDict):
    _root_path: List[str]

    def __init__(self, root_path: List[str]) -> None:
        super().__init__()
        self._root_path = root_path

    def __missing__(self, key):
        return ConfigItemWrapper(None, [*self._root_path, key])


class ConfigProvider(Protocol):
    """
    Config provider.

    Provides a list of ConfigurationReference.
    """

    def config(self) -> List[ConfigurationReference]:
        """Implement to provide a list of configuration items, for instance from Benchling APIs."""
        pass


class BenchlingConfigProvider(ConfigProvider):
    """
    Benchling Config provider.

    Provides a BenchlingAppConfiguration retrieved from Benchling's API.
    """

    _app_id: str
    _benchling: Benchling

    def __init__(self, benchling: Benchling, app_id: str):
        """
        Initialize Benchling Config Provider.

        :param benchling: A Benchling instance.
        :param app_id: The app_id from which to retrieve configuration.
        """
        self._app_id = app_id
        self._benchling = benchling

    def config(self) -> List[ConfigurationReference]:
        """Provide a Benchling app configuration from Benchling's APIs."""
        app_pages = self._benchling.apps.list_app_configuration_items(
            app_id=self._app_id,
            page_size=100,
            sort=ListAppConfigurationItemsSort.CREATEDATASC,
        )

        # Eager load all config items for now since we don't yet have a way of lazily querying by path
        all_config_pages = list(app_pages)
        # Punt on UnknownType for now as apps using manifests with new types +
        # older client could lead to unpredictable results
        all_config_items = [
            _supported_config_item(config_item) for page in all_config_pages for config_item in page
        ]

        return all_config_items


class StaticConfigProvider(ConfigProvider):
    """
    Static Config provider.

    Provides a BenchlingAppConfiguration from a static declaration. Useful for mocking or testing.
    """

    _configuration_items: List[ConfigurationReference]

    def __init__(self, configuration_items: List[ConfigurationReference]):
        """
        Initialize Static Config Provider.

        :param configuration_items: The configuration items to return.
        """
        self._configuration_items = configuration_items

    def config(self) -> List[ConfigurationReference]:
        """Provide Benchling app configuration items from a static reference."""
        return self._configuration_items


class ConfigItemWrapper:
    """
    Config Item Wrapper.

    A decorator class for AppConfigItem to assist with typesafe access to its values.
    Access the `item` attribute for the original config item, if present.
    """

    item: Optional[ConfigurationReference]
    path: List[str]

    def __init__(self, item: Optional[ConfigurationReference], path: List[str]) -> None:
        """Init Pathed Config Item."""
        self.item = item
        self.path = path

    def required(self) -> RequiredConfigItemWrapper:
        """Return a `RequiredPathedConfigItem` to enforce that config item is not optional."""
        if self.item is None:
            raise MissingRequiredConfigItemError(
                f"Required config item {self.path} is missing from the config store"
            )
        return RequiredConfigItemWrapper(self.item, self.path)

    def linked_resource(self) -> Optional[LinkedAppConfigResourceSummary]:
        """
        Return an optional LinkedAppConfigResourceSummary.

        Raises exceptions if the config item is not of the type that supports linked resources,
        or the linked resource is inaccessible.
        """
        if self.item is not None:
            # Python can't compare isinstance for Union types until Python 3.10 so just do this for now
            # from: https://github.com/python/typing/discussions/1132#discussioncomment-2560441
            # elif isinstance(self.item, get_args(LinkedResourceConfigurationReference)):
            if isinstance(
                self.item, (EntitySchemaAppConfigItem, FieldAppConfigItem, GenericApiIdentifiedAppConfigItem)
            ):
                if self.item.linked_resource is None:
                    return None
                elif isinstance(self.item.linked_resource, LinkedAppConfigResourceSummary):
                    return self.item.linked_resource
                elif isinstance(self.item.linked_resource, InaccessibleResource):
                    raise InaccessibleConfigItemError(
                        f"The linked resource for config item {self.path} was inaccessible. "
                        f"The caller does not have permissions to the resource."
                    )
                raise UnsupportedConfigItemError(
                    f"Unable to read app configuration with unsupported type: {self.item}"
                )
            raise ConfigItemLinkedResourceError(
                f"Type mismatch: The config item {self.item} type never has a linked resource"
            )
        return None

    def value(self) -> Union[str, float, int, bool, None]:
        """Return the value of the config item, if present."""
        return self.item.value if self.item else None

    def value_str(self) -> Optional[str]:
        """Return the value of the config item as a string, if present."""
        return str(self.value()) if self.value() else None

    def __bool__(self) -> bool:
        return self.value() is not None


class RequiredConfigItemWrapper(ConfigItemWrapper):
    """
    Required Config Item Wrapper.

    A decorator class for AppConfigItem to assist with typesafe access to its values.
    Enforces that a config item is present, and that it's value is not None.
    Access the `item` attribute for the original config item.
    """

    item: ConfigurationReference

    def __init__(self, item: ConfigurationReference, path: List[str]) -> None:
        """Init Required Pathed Config Item."""
        super().__init__(item, path)

    def linked_resource(self) -> LinkedAppConfigResourceSummary:
        """
        Return a LinkedAppConfigResourceSummary.

        Raises exceptions if the config item is not of the type that supports linked resources,
        or the linked resource is inaccessible.
        """
        linked_resource = super().linked_resource()
        if linked_resource is None:
            raise MissingRequiredConfigItemError(
                f"Required config item {self.path} is missing a linked resource"
            )
        return linked_resource

    def value(self) -> Union[str, float, int, bool]:
        """Return the value of the config item."""
        if self.item.value is None:
            raise MissingRequiredConfigItemError(
                f"Required config item {self.path} is missing from the config store"
            )
        return self.item.value

    def value_str(self) -> str:
        """Return the value of the config item as a string."""
        return str(self.value())


class ConfigItemStore:
    """
    Dependency Link Store.

    Marshals an app configuration from the configuration provider into an indexed structure.
    Only retrieves app configuration once unless its cache is invalidated.
    """

    _configuration_provider: ConfigProvider
    _configuration: Optional[List[ConfigurationReference]] = None
    _configuration_dict: Optional[Dict[ConfigItemPath, ConfigItemWrapper]] = None
    _array_path_row_names: Dict[Tuple[str, ...], OrderedSet[str]]

    def __init__(self, configuration_provider: ConfigProvider):
        """
        Initialize Dependency Link Store.

        :param configuration_provider: A ConfigProvider that will be invoked to provide the
        underlying config from which to organize dependency links.
        """
        self._configuration_provider = configuration_provider
        self._array_path_row_names = dict()

    @property
    def configuration(self) -> List[ConfigurationReference]:
        """
        Get the underlying configuration.

        Return the raw, stored configuration. Can be used if the provided accessors are inadequate
        to find particular configuration items.
        """
        if not self._configuration:
            self._configuration = self._configuration_provider.config()
        return self._configuration

    @property
    def configuration_path_dict(self) -> Dict[ConfigItemPath, ConfigItemWrapper]:
        """
        Config links.

        Return a dict of configuration item paths to their corresponding configuration items.
        """
        if not self._configuration_dict:
            self._configuration_dict = {
                tuple(item.path): ConfigItemWrapper(item, item.path) for item in self.configuration
            }
        return self._configuration_dict

    def config_by_path(self, path: List[str]) -> ConfigItemWrapper:
        """
        Config by path.

        Find an app config item by its exact path match, if it exists. Does not search partial paths.
        """
        # Since we eager load all config now, we know that missing path means it's not configured in Benchling
        # Later if we support lazy loading, we'll need to differentiate what's in our cache versus missing
        return self.configuration_path_dict.get(tuple(path), ConfigItemWrapper(None, path))

    def config_keys_by_path(self, path: List[str]) -> OrderedSet[str]:
        """
        Config keys by path.

        Find a set of app config keys at the specified path, if any. Does not return keys that are nested
        beyond the current level.

        For instance, given paths:
        ["One", "Two"]
        ["One", "Two", "Three"]
        ["One", "Two", "Four"]
        ["One", "Two", "Three", "Five"]
        ["Zero", "One", "Two", "Three"]

        The expected return from this method when path=["One", "Two"] is a set {"Three", "Four"}.
        """
        # Convert path to tuple, as list is not hashable for dict keys
        path_tuple = tuple(path)
        if path_tuple not in self._array_path_row_names:
            self._array_path_row_names[path_tuple] = OrderedSet(
                [
                    config_item.path[len(path)]
                    # Use the list instead of configuration_dict to preserve order
                    for config_item in self.configuration
                    # The +1 is the name of the array row
                    if len(config_item.path) >= len(path) + 1
                    # Ignoring flake8 error E203 because black keeps putting in whitespace padding :
                    and config_item.path[0 : len(path_tuple)] == path
                    and config_item.value is not None
                ]
            )
        return self._array_path_row_names[path_tuple]

    def array_rows_to_dict(self, path: List[str]) -> OrderedDict[str, Dict[str, ConfigItemWrapper]]:
        """Given a path to the root of a config array, return each element as a named dict."""
        array_keys = self.config_keys_by_path(path)
        # Although we don't have a way of preserving order when pulling array elements from the API right now
        # we should intentionally order these to accommodate a potential ordered future
        array_elements_map = _DefaultOrderedDict(root_path=path)
        for key in array_keys:
            array_elements_map[key] = _DefaultOrderedDict(root_path=[*path, key])
            # Don't care about order for the keys within a row, only the order of the rows themselves
            for array_element_key in self.config_keys_by_path([*path, key]):
                if self.config_by_path([*path, key, array_element_key]) is not None:
                    array_elements_map[key][array_element_key] = self.config_by_path(
                        [*path, key, array_element_key]
                    )
        return array_elements_map

    def invalidate_cache(self) -> None:
        """
        Invalidate Cache.

        Will force retrieval of configuration from the ConfigProvider the next time the link store is accessed.
        """
        self._configuration = None
        self._configuration_dict = None
        self._array_path_row_names = dict()


def _supported_config_item(config_item: AppConfigItem) -> ConfigurationReference:
    if isinstance(config_item, UnknownType):
        raise UnsupportedConfigItemError(
            f"Unable to read app configuration with unsupported type: {config_item}"
        )
    return config_item



from typing import Iterable, List, Optional

from benchling_api_client.v2.stable.api.assay_runs import (
    archive_assay_runs,
    bulk_get_assay_runs,
    create_assay_runs,
    get_assay_run,
    list_assay_runs,
    list_automation_input_generators,
    list_automation_output_processors_deprecated,
    unarchive_assay_runs,
    update_assay_run,
)
from benchling_api_client.v2.types import Response

from benchling_sdk.errors import raise_for_status
from benchling_sdk.helpers.decorators import api_method
from benchling_sdk.helpers.logging_helpers import log_deprecation
from benchling_sdk.helpers.pagination_helpers import NextToken, PageIterator
from benchling_sdk.helpers.response_helpers import model_from_detailed
from benchling_sdk.helpers.serialization_helpers import (
    array_query_param,
    none_as_unset,
    optional_array_query_param,
)
from benchling_sdk.models import (
    AssayRun,
    AssayRunCreate,
    AssayRunsArchivalChange,
    AssayRunsArchive,
    AssayRunsArchiveReason,
    AssayRunsBulkCreateRequest,
    AssayRunsBulkCreateResponse,
    AssayRunsPaginatedList,
    AssayRunsUnarchive,
    AssayRunUpdate,
    AutomationFileInputsPaginatedList,
    AutomationInputGenerator,
    AutomationOutputProcessor,
    AutomationOutputProcessorsPaginatedList,
)
from benchling_sdk.services.v2.base_service import BaseService


class AssayRunService(BaseService):
    """
    Assay Runs.

    Runs capture the details / parameters of a run that was performed. Results are usually nested under a run.

    See https://benchling.com/api/reference#/Assay%20Runs
    """

    @api_method
    def get_by_id(self, assay_run_id: str) -> AssayRun:
        """
        Get a run.

        See https://benchling.com/api/reference#/Assay%20Runs/getAssayRun
        """
        response = get_assay_run.sync_detailed(client=self.client, assay_run_id=assay_run_id)
        return model_from_detailed(response)

    @api_method
    def _assay_runs_page(
        self,
        schema_id: str,
        min_created_time: Optional[int] = None,
        max_created_time: Optional[int] = None,
        next_token: Optional[NextToken] = None,
        page_size: Optional[int] = None,
        ids: Optional[Iterable[str]] = None,
    ) -> Response[AssayRunsPaginatedList]:
        response = list_assay_runs.sync_detailed(
            client=self.client,
            schema_id=schema_id,
            min_created_time=none_as_unset(min_created_time),
            max_created_time=none_as_unset(max_created_time),
            ids=none_as_unset(optional_array_query_param(ids)),
            next_token=none_as_unset(next_token),
            page_size=none_as_unset(page_size),
        )
        raise_for_status(response)
        return response

    def list(
        self,
        schema_id: str,
        min_created_time: Optional[int] = None,
        max_created_time: Optional[int] = None,
        page_size: Optional[int] = None,
        ids: Optional[Iterable[str]] = None,
    ) -> PageIterator[AssayRun]:
        """
        List runs.

        See https://benchling.com/api/reference#/Assay%20Runs/listAssayRuns
        """

        def api_call(next_token: NextToken) -> Response[AssayRunsPaginatedList]:
            return self._assay_runs_page(
                schema_id=schema_id,
                min_created_time=min_created_time,
                max_created_time=max_created_time,
                ids=ids,
                next_token=next_token,
                page_size=page_size,
            )

        def results_extractor(body: AssayRunsPaginatedList) -> Optional[List[AssayRun]]:
            return body.assay_runs

        return PageIterator(api_call, results_extractor)

    @api_method
    def bulk_get(self, assay_run_ids: Iterable[str]) -> Optional[List[AssayRun]]:
        """
        Bulk get runs by ID.

        See https://benchling.com/api/reference#/Assay%20Runs/bulkGetAssayRuns
        """
        run_ids_string = array_query_param(assay_run_ids)
        response = bulk_get_assay_runs.sync_detailed(client=self.client, assay_run_ids=run_ids_string)
        runs_list = model_from_detailed(response)
        return runs_list.assay_runs

    @api_method
    def create(self, assay_runs: Iterable[AssayRunCreate]) -> AssayRunsBulkCreateResponse:
        """
        Create 1 or more runs.

        See https://benchling.com/api/reference#/Assay%20Runs/createAssayRuns
        """
        create_runs = AssayRunsBulkCreateRequest(assay_runs=list(assay_runs))
        response = create_assay_runs.sync_detailed(client=self.client, json_body=create_runs)
        return model_from_detailed(response)

    @api_method
    def update(self, assay_run_id: str, assay_run: AssayRunUpdate) -> AssayRun:
        """
        Update a run.

        See https://benchling.com/api/reference#/Assay%20Runs/updateAssayRun
        """
        response = update_assay_run.sync_detailed(
            client=self.client, assay_run_id=assay_run_id, json_body=assay_run
        )
        return model_from_detailed(response)

    @api_method
    def _automation_input_generators_page(
        self,
        assay_run_id: str,
        modified_at: Optional[str] = None,
        next_token: Optional[NextToken] = None,
    ) -> Response[AutomationFileInputsPaginatedList]:
        response = list_automation_input_generators.sync_detailed(
            client=self.client,
            assay_run_id=assay_run_id,
            modified_at=none_as_unset(modified_at),
            next_token=none_as_unset(next_token),
        )
        raise_for_status(response)
        return response  # type: ignore

    def automation_input_generators(
        self, assay_run_id: str, modified_at: Optional[str] = None
    ) -> PageIterator[AutomationInputGenerator]:
        """
        List AutomationInputGenerators by Run.

        See https://benchling.com/api/reference#/Assay%20Runs/listAutomationInputGenerators
        """

        def api_call(next_token: NextToken) -> Response[AutomationFileInputsPaginatedList]:
            return self._automation_input_generators_page(
                assay_run_id=assay_run_id, modified_at=modified_at, next_token=next_token
            )

        def results_extractor(body: AutomationFileInputsPaginatedList) -> List[AutomationInputGenerator]:
            return body.automation_input_generators

        return PageIterator(api_call, results_extractor)

    @api_method
    def _automation_output_processors_page(
        self,
        assay_run_id: str,
        next_token: Optional[NextToken] = None,
    ) -> Response[AutomationOutputProcessorsPaginatedList]:
        """Deprecated in favor of lab_automation.automation_output_processors."""
        log_deprecation(
            "assay_runs.automation_output_processors_page", "lab_automation.automation_output_processors"
        )
        response = list_automation_output_processors_deprecated.sync_detailed(
            client=self.client,
            assay_run_id=assay_run_id,
            next_token=none_as_unset(next_token),
        )
        raise_for_status(response)
        return response  # type: ignore

    def automation_output_processors(self, assay_run_id: str) -> PageIterator[AutomationOutputProcessor]:
        """
        List AutomationOutputProcessors by Run.

        Deprecated in favor of lab_automation.automation_output_processors.

        See https://benchling.com/api/reference#/Assay%20Runs/listAutomationOutputProcessorsDeprecated
        """
        log_deprecation(
            "assay_runs.automation_output_processors", "lab_automation.automation_output_processors"
        )

        def api_call(next_token: NextToken) -> Response[AutomationOutputProcessorsPaginatedList]:
            return self._automation_output_processors_page(assay_run_id=assay_run_id, next_token=next_token)

        def results_extractor(
            body: AutomationOutputProcessorsPaginatedList,
        ) -> List[AutomationOutputProcessor]:
            return body.automation_output_processors

        return PageIterator(api_call, results_extractor)

    @api_method
    def archive(
        self, assay_run_ids: Iterable[str], reason: AssayRunsArchiveReason
    ) -> AssayRunsArchivalChange:
        """
        Archive Assay Runs.

        See https://benchling.com/api/reference#/Assay%20Runs/archiveAssayRuns
        """
        archive_request = AssayRunsArchive(assay_run_ids=list(assay_run_ids), reason=reason)
        response = archive_assay_runs.sync_detailed(client=self.client, json_body=archive_request)
        return model_from_detailed(response)

    @api_method
    def unarchive(self, assay_run_ids: Iterable[str]) -> AssayRunsArchivalChange:
        """
        Unarchive Assay Runs.

        See https://benchling.com/api/reference#/Assay%20Runs/unarchiveAssayRuns
        """
        unarchive_request = AssayRunsUnarchive(assay_run_ids=list(assay_run_ids))
        response = unarchive_assay_runs.sync_detailed(client=self.client, json_body=unarchive_request)
        return model_from_detailed(response)



from typing import Iterable, List, Optional

from benchling_api_client.v2.stable.api.test_orders import (
    bulk_update_test_orders,
    list_test_orders,
    update_test_order,
)
from benchling_api_client.v2.stable.models.async_task_link import AsyncTaskLink
from benchling_api_client.v2.stable.models.list_test_orders_sort import ListTestOrdersSort
from benchling_api_client.v2.stable.models.test_order import TestOrder
from benchling_api_client.v2.stable.models.test_order_bulk_update import TestOrderBulkUpdate
from benchling_api_client.v2.stable.models.test_order_status import TestOrderStatus
from benchling_api_client.v2.stable.models.test_order_update import TestOrderUpdate
from benchling_api_client.v2.stable.models.test_orders_bulk_update_request import TestOrdersBulkUpdateRequest
from benchling_api_client.v2.stable.models.test_orders_paginated_list import TestOrdersPaginatedList
from benchling_api_client.v2.types import Response

from benchling_sdk.errors import raise_for_status
from benchling_sdk.helpers.decorators import api_method
from benchling_sdk.helpers.pagination_helpers import NextToken, PageIterator
from benchling_sdk.helpers.response_helpers import model_from_detailed
from benchling_sdk.helpers.serialization_helpers import none_as_unset
from benchling_sdk.services.v2.base_service import BaseService


class TestOrderService(BaseService):
    """
    Test Orders.

    Test orders enable users to order tests for specific sample/container combinations that will be fulfilled in assays.

    See https://benchling.com/api/reference?availability=la#/Test%20Orders/
    """

    @api_method
    def bulk_update(self, test_orders: Iterable[TestOrderBulkUpdate]) -> AsyncTaskLink:
        """
        Bulk update Test Orders.

        See https://benchling.com/api/reference?availability=la#/Test%20Orders/bulkUpdateTestOrders
        """
        body = TestOrdersBulkUpdateRequest(list(test_orders))
        response = bulk_update_test_orders.sync_detailed(client=self.client, json_body=body)
        return model_from_detailed(response)

    @api_method
    def _test_orders_page(
        self,
        page_size: Optional[int] = 50,
        next_token: Optional[str] = None,
        sort: Optional[ListTestOrdersSort] = ListTestOrdersSort.MODIFIEDATDESC,
        created_atlt: Optional[str] = None,
        created_atgt: Optional[str] = None,
        created_atlte: Optional[str] = None,
        created_atgte: Optional[str] = None,
        modified_atlt: Optional[str] = None,
        modified_atgt: Optional[str] = None,
        modified_atlte: Optional[str] = None,
        modified_atgte: Optional[str] = None,
        ids: Optional[str] = None,
        container_idsany_of: Optional[str] = None,
        sample_idsany_of: Optional[str] = None,
        status: Optional[TestOrderStatus] = None,
    ) -> Response[TestOrdersPaginatedList]:
        response = list_test_orders.sync_detailed(
            client=self.client,
            page_size=none_as_unset(page_size),
            next_token=none_as_unset(next_token),
            sort=none_as_unset(sort),
            created_atlt=none_as_unset(created_atlt),
            created_atgt=none_as_unset(created_atgt),
            created_atlte=none_as_unset(created_atlte),
            created_atgte=none_as_unset(created_atgte),
            modified_atlt=none_as_unset(modified_atlt),
            modified_atgt=none_as_unset(modified_atgt),
            modified_atlte=none_as_unset(modified_atlte),
            modified_atgte=none_as_unset(modified_atgte),
            ids=none_as_unset(ids),
            container_idsany_of=none_as_unset(container_idsany_of),
            sample_idsany_of=none_as_unset(sample_idsany_of),
            status=none_as_unset(status),
        )
        raise_for_status(response)
        return response  # type: ignore

    def list(
        self,
        *,
        page_size: Optional[int] = 50,
        next_token: Optional[str] = None,
        sort: Optional[ListTestOrdersSort] = ListTestOrdersSort.MODIFIEDATDESC,
        created_atlt: Optional[str] = None,
        created_atgt: Optional[str] = None,
        created_atlte: Optional[str] = None,
        created_atgte: Optional[str] = None,
        modified_atlt: Optional[str] = None,
        modified_atgt: Optional[str] = None,
        modified_atlte: Optional[str] = None,
        modified_atgte: Optional[str] = None,
        ids: Optional[str] = None,
        container_idsany_of: Optional[str] = None,
        sample_idsany_of: Optional[str] = None,
        status: Optional[TestOrderStatus] = None,
    ) -> PageIterator[TestOrder]:
        """
        List Test Orders.

        See https://benchling.com/api/reference?availability=la#/Test%20Orders/listTestOrders
        """

        def api_call(next_token: NextToken) -> Response[TestOrdersPaginatedList]:
            return self._test_orders_page(
                page_size=page_size,
                next_token=next_token,
                sort=sort,
                created_atlt=created_atlt,
                created_atgt=created_atgt,
                created_atlte=created_atlte,
                created_atgte=created_atgte,
                modified_atlt=modified_atlt,
                modified_atgt=modified_atgt,
                modified_atlte=modified_atlte,
                modified_atgte=modified_atgte,
                ids=ids,
                container_idsany_of=container_idsany_of,
                sample_idsany_of=sample_idsany_of,
                status=status,
            )

        def results_extractor(body: TestOrdersPaginatedList) -> Optional[List[TestOrder]]:
            return body.test_orders

        return PageIterator(api_call, results_extractor)

    @api_method
    def update(self, test_order_id: str, test_order: TestOrderUpdate) -> TestOrder:
        """
        Update a TestOrder.

        See https://benchling.com/api/reference?availability=la#/Test%20Orders/updateTestOrder
        """
        response = update_test_order.sync_detailed(
            client=self.client, test_order_id=test_order_id, json_body=test_order
        )
        return model_from_detailed(response)



from benchling_api_client.v2.stable.api.instrument_queries import get_instrument_query

from benchling_sdk.helpers.decorators import api_method
from benchling_sdk.helpers.response_helpers import model_from_detailed
from benchling_sdk.models import InstrumentQuery
from benchling_sdk.services.v2.base_service import BaseService


class InstrumentQueryService(BaseService):
    """
    Instrument Queries.

    Instrument Queries are used to query the instrument service.

    See https://benchling.com/api/reference#/Instrument%20Queries
    """

    @api_method
    def get_by_id(self, instrument_query_id: str) -> InstrumentQuery:
        """
        Get an instrument query.

        See https://benchling.com/api/reference#/Instrument%20Queries/getInstrumentQuery
        """
        response = get_instrument_query.sync_detailed(
            client=self.client, instrument_query_id=instrument_query_id
        )
        return model_from_detailed(response)



from dataclasses import dataclass
from typing import Any, cast, Generic, List, Optional, Type, TypeVar, Union

from benchling_api_client.v2.stable.client import Client

from benchling_sdk.helpers.serialization_helpers import unset_as_none
from benchling_sdk.models import AsyncTaskErrors, AsyncTaskErrorsItem, AsyncTaskLink, AsyncTaskStatus

ResponseT = TypeVar("ResponseT")


@dataclass
class TaskCompletion(Generic[ResponseT]):
    """Return type for TaskHelper.wait_for_task, same as AsyncTask but with a typed response."""

    success: bool
    errors: Optional[Union[AsyncTaskErrors, List[AsyncTaskErrorsItem]]] = None
    message: Optional[str] = None
    response: Optional[ResponseT] = None


@dataclass
class TaskFailureException(Exception):
    """Exception type used by :py:class:`.TaskHelper` methods."""

    errors: Union[AsyncTaskErrors, List[AsyncTaskErrorsItem]]
    message: Optional[str]


class EmptyTaskResponse:
    """A sentinel object used for tasks that do not return any response data on completion."""

    pass


EMPTY_TASK_RESPONSE = EmptyTaskResponse()


class TaskHelper(AsyncTaskLink, Generic[ResponseT]):
    """
    Used by Benchling async task endpoints to provide the task response in an appropriate type.

    In the API spec, endpoints that create a long-running task are defined as returning an
    :py:class:`benchling_sdk.models.AsyncTaskLink`, which can be used with
    :py:class:`benchling_sdk.services.v2.stable.TaskService` to query the task status and
    response. But AsyncTaskLink and the task query endpoint do not define a specific schema
    for the task response.

    To work around that limitation, those SDK endpoints now return a TaskHelper instead. This is
    subclassed from AsyncTaskLink for backward compatibility, but unlike AsyncTaskLink, TaskHelper
    knows what the type of the task response should be. It also retains an association with the API
    client, so rather than calling a separate service method, you can simply call
    :py:meth:`.TaskHelper.wait_for_completion` or :py:meth:`.TaskHelper.wait_for_response`.

    You can access a task for up to 30 minutes after its completion, after which its data will no
    longer be available.
    """

    _client: Client
    _response_class: Type[ResponseT]
    _response_decoder: Any
    # _response_decoder is really a Callable, but adding that type hint would cause problems because
    # mypy would treat it as an instance method and assume it should have a self parameter.

    def __init__(self, from_task_link: AsyncTaskLink, client: Client, response_class: Type[ResponseT]):
        """Initialize the instance. This should only be used internally by Benchling API methods."""
        self.task_id = from_task_link.task_id
        self.additional_properties = from_task_link.additional_properties
        self._client = client
        self._response_class = response_class

        if response_class is not EmptyTaskResponse:
            assert hasattr(response_class, "from_dict")
            self._response_decoder = getattr(response_class, "from_dict")  # noqa: B009

    def wait_for_completion(
        self, interval_wait_seconds: int = 1, max_wait_seconds: int = 600
    ) -> TaskCompletion[ResponseT]:
        """
        Wait for the task to succeed or fail.

        This is equivalent to the :py:meth:`benchling_sdk.services.v2.stable.task_service.TaskService.wait_for_task`
        method in :py:class:`benchling_sdk.services.v2.stable.task_service.TaskService`, except that
        instead of returning an :py:class:`benchling_sdk.models.AsyncTask` whose `response` property is
        a dict, it returns a :py:class:`TaskCompletion` whose `response` property is the appropriate type
        for the API method you called. For instance, if the method was `AaSequenceService.bulk_create`,
        the response type will be :py:class:`benchling_sdk.models.BulkCreateAaSequencesAsyncTaskResponse`.

        :param interval_wait_seconds: time to wait between API calls in seconds
        :param max_wait_seconds: maximum wait time in seconds before raising an error
        :return: The task completion status. Check `status` for success or failure
        :rtype: TaskCompletion
        :raises benchling_sdk.errors.WaitForTaskExpiredError: if the maximum wait time has elapsed
        """
        from benchling_sdk.helpers.client_helpers import v2_stable_client
        from benchling_sdk.services.v2.stable.task_service import TaskService

        # The "get task" polling endpoint is part of the v2 stable API; it doesn't exist in alpha or beta.
        task_service = TaskService(v2_stable_client(self._client))

        task = task_service.wait_for_task(
            self.task_id, interval_wait_seconds=interval_wait_seconds, max_wait_seconds=max_wait_seconds
        )
        response: Optional[ResponseT]
        if task.status != AsyncTaskStatus.SUCCEEDED:
            response = None
        elif self._response_class is EmptyTaskResponse:
            response = cast(ResponseT, EMPTY_TASK_RESPONSE)
        else:
            response = self._response_decoder(task.response.to_dict())
        errors = None if task.status != AsyncTaskStatus.FAILED else unset_as_none(lambda: task.errors)
        message = None if task.status != AsyncTaskStatus.FAILED else unset_as_none(lambda:task.message)
        return TaskCompletion(
            success=task.status == AsyncTaskStatus.SUCCEEDED,
            errors=errors,  # type: ignore
            message=message,
            response=response,
        )

    def wait_for_response(self, interval_wait_seconds: int = 1, max_wait_seconds: int = 600) -> ResponseT:
        """
        Wait for the task and return the response object on success, or raise an exception on failure.

        This is a convenience method for calling :py:meth:`wait_for_completion` and then getting the
        `response` property of the returned object if the task succeeded, in cases where you're not
        interested in any :py:class:`.TaskCompletion` properties except the response.

        The type of the returned object is depends on the API method you called. For instance, if the
        method was `AaSequenceService.bulk_create`, the response type will be
        :py:class:`benchling_sdk.models.BulkCreateAaSequencesAsyncTaskResponse`.

        :param interval_wait_seconds: time to wait between API calls in seconds
        :param max_wait_seconds: maximum wait time in seconds before raising an error
        :return: The response object, if the task succeeded.
        :rtype: ResponseT
        :raises benchling_sdk.errors.WaitForTaskExpiredError: if the maximum wait time has elapsed
        :raises .TaskFailureException: if the task failed
        """
        task_completion = self.wait_for_completion(interval_wait_seconds, max_wait_seconds)
        if not task_completion.success:
            raise TaskFailureException(task_completion.errors or AsyncTaskErrors(), task_completion.message)
        assert task_completion.response, "completed task should have contained a response"
        # This assertion matches the behavior of TaskHelper.wait_for_completion(): response is only set
        # to None if the task failed, otherwise it should always be an object even if that object is an
        # EmptyTaskResponse. This also matches how we define async task models in the API spec, where
        # if we don't care about the response field, it's an empty object rather than null.

        return task_completion.response



from abc import ABC, abstractmethod

from benchling_sdk.apps.config.cryptography_helpers import _create_key_type
from benchling_sdk.helpers.package_helpers import _required_packages_context, ExtrasPackage


class BaseDecryptionProvider(ABC):
    """
    Provides a way to decrypt encrypted messages.

    Various implementations might use AWS KMS, Azure, etc.
    """

    @abstractmethod
    def decrypt(self, ciphertext: str) -> str:
        """
        Decrypt secrets.

        Receives encrypted cipher text and returns the decrypted plain text.
        """
        pass


class BaseKeyUnwrappingDecryptionProvider(BaseDecryptionProvider):
    """
    A decryption provider that will unwrap the key received from Benchling.

    Generally prefer extending this to BaseDecryptionProvider, unless you
    need to unwrap the key received from Benchling yourself.
    """

    _charset: str

    @_required_packages_context(ExtrasPackage.PYTHON_JOSE)
    def __init__(self, charset: str = "utf-8"):
        """Init BaseKeyUnwrappingDecryptionProvider."""
        self._charset = charset
        from jose import jwk
        from jose.constants import ALGORITHMS

        def _unwrapping_function(wrapped_key: bytes) -> bytes:
            return self.unwrap_key(wrapped_key)

        jwk.register_key(ALGORITHMS.RSA_OAEP_256, _create_key_type(_unwrapping_function))

    @_required_packages_context(ExtrasPackage.PYTHON_JOSE)
    def decrypt(self, ciphertext: str) -> str:
        """
        Decrypt.

        Implement BaseDecryptionProvider's decrypt. Unwraps the encrypted payload from
        Benchling and will delegate to unwrap_key() to fetch the key decrypting key to
        then decrypt the secret.
        """
        from jose import jwe

        # Since we abstract key creation, the key name doesn't really matter
        return jwe.decrypt(ciphertext, "unused_key_id").decode(self._charset)

    @abstractmethod
    def unwrap_key(self, wrapped_key: bytes) -> bytes:
        """
        Unwrap Key.

        Accepts the wrapped key and decrypts the unwrapped key to use for data decryption
        from a provider such as AWS KMS, Azure, etc.
        """
        pass



# Tuple instead of list so it's hashable and preserves order
from typing import Tuple, Union

from benchling_sdk.models import (
    ArrayElementAppConfigItem,
    BooleanAppConfigItem,
    DateAppConfigItem,
    DatetimeAppConfigItem,
    EntitySchemaAppConfigItem,
    FieldAppConfigItem,
    FloatAppConfigItem,
    GenericApiIdentifiedAppConfigItem,
    IntegerAppConfigItem,
    JsonAppConfigItem,
    SecureTextAppConfigItem,
    TextAppConfigItem,
)

# Tuple for an ordered collection that's hashable
ConfigItemPath = Tuple[str, ...]

# Everything from AppConfigItem except UnknownType
ConfigurationReference = Union[
    ArrayElementAppConfigItem,
    DateAppConfigItem,
    DatetimeAppConfigItem,
    JsonAppConfigItem,
    EntitySchemaAppConfigItem,
    FieldAppConfigItem,
    BooleanAppConfigItem,
    IntegerAppConfigItem,
    FloatAppConfigItem,
    GenericApiIdentifiedAppConfigItem,
    SecureTextAppConfigItem,
    TextAppConfigItem,
]



from __future__ import annotations

from functools import cached_property
from typing import Optional, TYPE_CHECKING

from benchling_api_client.v2.stable.client import Client

from benchling_sdk.helpers.client_helpers import v2_beta_client
from benchling_sdk.helpers.retry_helpers import RetryStrategy
from benchling_sdk.services.v2.base_service import BaseService
from benchling_sdk.services.v2.beta.v2_beta_data_frame_service import V2BetaDataFrameService

if TYPE_CHECKING:
    from benchling_sdk.services.v2.beta.v2_beta_analysis_service import V2BetaAnalysisService
    from benchling_sdk.services.v2.beta.v2_beta_app_definition_service import V2BetaAppDefinitionService
    from benchling_sdk.services.v2.beta.v2_beta_app_service import V2BetaAppService
    from benchling_sdk.services.v2.beta.v2_beta_collaboration_service import V2BetaCollaborationService
    from benchling_sdk.services.v2.beta.v2_beta_entry_service import V2BetaEntryService
    from benchling_sdk.services.v2.beta.v2_beta_folder_service import V2BetaFolderService
    from benchling_sdk.services.v2.beta.v2_beta_project_service import V2BetaProjectService
    from benchling_sdk.services.v2.beta.v2_beta_worklist_service import V2BetaWorklistService


class V2BetaService(BaseService):
    """
    V2-beta.

    Beta endpoints have different stability guidelines than other stable endpoints.

    See https://benchling.com/api/v2-beta/reference
    """

    _beta_client: Client

    def __init__(self, client: Client, retry_strategy: Optional[RetryStrategy] = None):
        """
        Initialize a v2-beta service.

        :param client: Underlying generated Client.
        :param retry_strategy: Retry strategy for failed HTTP calls
        """
        super().__init__(client, retry_strategy)
        self._beta_client = v2_beta_client(self.client)

    @cached_property
    def analyses(self) -> V2BetaAnalysisService:
        """
        V2-Beta Analyses.

        Analyses allow experimental data to be viewed, analyzed, and visualized.

        https://benchling.com/api/v2-beta/reference#/Analyses
        """
        from .beta.v2_beta_analysis_service import V2BetaAnalysisService

        return self._create_service(V2BetaAnalysisService)

    @cached_property
    def apps(self) -> V2BetaAppService:
        """
        V2-Beta Apps.

        Create and manage Apps on your tenant.

        https://benchling.com/api/v2-beta/reference#/Apps
        """
        from .beta.v2_beta_app_service import V2BetaAppService

        return self._create_service(V2BetaAppService)

    @cached_property
    def app_definitions(self) -> V2BetaAppDefinitionService:
        """
        V2-Beta App Definitions.

        Create and manage Benchling app definitions on your tenant.

        https://benchling.com/api/v2-beta/reference#/App%20Definitions
        """
        from .beta.v2_beta_app_definition_service import V2BetaAppDefinitionService

        return self._create_service(V2BetaAppDefinitionService)

    @cached_property
    def collaborations(self) -> V2BetaCollaborationService:
        """
        V2-Beta Collaborations.

        Collaborations represent which user or group has which access policies.

        See https://benchling.com/api/v2-beta/reference?showLA=true#/Collaborations
        """
        from .beta.v2_beta_collaboration_service import V2BetaCollaborationService

        return self._create_service(V2BetaCollaborationService)

    @cached_property
    def data_frames(self) -> V2BetaDataFrameService:
        """
        V2-Beta DataFrames.

        DataFrames are Benchling objects that represent tabular data with typed columns and rows of data.

        See https://benchling.com/api/v2-beta/reference#/Data%20Frames
        """
        from .beta.v2_beta_data_frame_service import V2BetaDataFrameService

        return self._create_service(V2BetaDataFrameService)

    @cached_property
    def entries(self) -> V2BetaEntryService:
        """
        V2-Beta Entries.

        Entries are rich text documents that allow you to capture all of your experimental data in one place.

        https://benchling.com/api/v2-beta/reference#/Entries
        """
        from .beta.v2_beta_entry_service import V2BetaEntryService

        return self._create_service(V2BetaEntryService)

    @cached_property
    def folders(self) -> V2BetaFolderService:
        """
        V2-Beta Folders.

        Folders are nested within projects to provide additional organization.

        https://benchling.com/api/v2-beta/reference?showLA=true#/Folders
        """
        from .beta.v2_beta_folder_service import V2BetaFolderService

        return self._create_service(V2BetaFolderService)

    @cached_property
    def projects(self) -> V2BetaProjectService:
        """
        V2-Beta Projects.

        Manage project objects.

        See https://benchling.com/api/v2-beta/reference?#/Projects
        """
        from .beta.v2_beta_project_service import V2BetaProjectService

        return self._create_service(V2BetaProjectService)

    @cached_property
    def worklists(self) -> V2BetaWorklistService:
        """
        V2-Beta Worklists.

        Worklists are a convenient way to organize items for bulk actions, and are complementary to folders and
        projects.

        See https://benchling.com/api/v2-beta/reference#/Worklists
        """
        from .beta.v2_beta_worklist_service import V2BetaWorklistService

        return self._create_service(V2BetaWorklistService)

    def _create_service(self, cls):
        """Instantiate a service using the beta client."""
        return cls(self._beta_client, self._retry_strategy)



from __future__ import annotations

from typing import Optional

from benchling_sdk.apps.config.framework import BenchlingConfigProvider, ConfigItemStore
from benchling_sdk.apps.status.framework import (
    continue_session_context,
    new_session_context,
    SessionContextEnterHandler,
    SessionContextExitHandler,
    SessionContextManager,
)
from benchling_sdk.benchling import Benchling


class App:
    """
    App.

    See https://docs.benchling.com/docs/getting-started-benchling-apps

    Accepts providers as arguments to lazily initialize since some required attributes may not be
    known until runtime. Also allows for easier mocking in tests.
    """

    _app_id: str
    _benchling: Benchling
    _config_store: ConfigItemStore

    def __init__(
        self, app_id: str, benchling: Benchling, config_store: Optional[ConfigItemStore] = None
    ) -> None:
        """
        Initialize a Benchling App.

        :param app_id: An id representing a tenanted app installation (e.g., "app_Uh3BZ55aYcXGFJVb")
        :param benchling: A Benchling object for making API calls. The auth_method should be valid for the specified App.
            Commonly this is ClientCredentialsOAuth2 using the app's client ID and client secret.
        :param config_store: The configuration item store for accessing an App's tenanted app config items.
            If unspecified, will default to retrieving app config from the tenant referenced by Benchling.
            Apps that don't use app configuration can safely ignore this.
        """
        self._app_id = app_id
        self._benchling = benchling
        self._config_store = (
            config_store if config_store else ConfigItemStore(BenchlingConfigProvider(benchling, app_id))
        )

    @property
    def id(self) -> str:
        """Return the app tenanted installation id."""
        return self._app_id

    @property
    def benchling(self) -> Benchling:
        """Return a Benchling instance for the App."""
        return self._benchling

    @property
    def config_store(self) -> ConfigItemStore:
        """Return a ConfigItemStore instance for the App."""
        return self._config_store

    def create_session_context(
        self,
        name: str,
        timeout_seconds: int,
        context_enter_handler: Optional[SessionContextEnterHandler] = None,
        context_exit_handler: Optional[SessionContextExitHandler] = None,
    ) -> SessionContextManager:
        """
        Create Session Context.

        Create a new app session in Benchling.
        """
        return new_session_context(self, name, timeout_seconds, context_enter_handler, context_exit_handler)

    def continue_session_context(
        self,
        session_id: str,
        context_enter_handler: Optional[SessionContextEnterHandler] = None,
        context_exit_handler: Optional[SessionContextExitHandler] = None,
    ) -> SessionContextManager:
        """
        Continue Session Context.

        Fetch an existing app session from Benchling and enter a context with it.
        """
        return continue_session_context(self, session_id, context_enter_handler, context_exit_handler)



from typing import List, Optional

from benchling_api_client.v2.stable.api.events import list_events
from benchling_api_client.v2.types import Response

from benchling_sdk.errors import raise_for_status
from benchling_sdk.helpers.decorators import api_method
from benchling_sdk.helpers.pagination_helpers import NextToken, PageIterator
from benchling_sdk.helpers.serialization_helpers import none_as_unset
from benchling_sdk.models import Event, EventsPaginatedList
from benchling_sdk.services.v2.base_service import BaseService


class EventService(BaseService):
    """
    Events.

    The Events system allows external services to subscribe to events that are triggered in Benchling (e.g. plasmid
    registration, request submission, etc).

    See https://benchling.com/api/reference#/Events
    """

    @api_method
    def _events_page(
        self,
        created_atgte: Optional[str] = None,
        starting_after: Optional[str] = None,
        event_types: Optional[str] = None,
        poll: Optional[bool] = None,
        page_size: Optional[int] = None,
        next_token: Optional[NextToken] = None,
    ) -> Response[EventsPaginatedList]:
        response = list_events.sync_detailed(
            client=self.client,
            starting_after=none_as_unset(starting_after),
            created_atgte=none_as_unset(created_atgte),
            event_types=none_as_unset(event_types),
            poll=none_as_unset(poll),
            page_size=none_as_unset(page_size),
            next_token=none_as_unset(next_token),
        )
        raise_for_status(response)
        return response  # type: ignore

    def list(
        self,
        created_atgte: Optional[str] = None,
        starting_after: Optional[str] = None,
        event_types: Optional[str] = None,
        poll: Optional[bool] = None,
        page_size: Optional[int] = None,
    ) -> PageIterator[Event]:
        """
        List Events.

        See https://benchling.com/api/reference#/Events/listEvents
        """

        def api_call(next_token: NextToken) -> Response[EventsPaginatedList]:
            return self._events_page(
                starting_after=starting_after,
                created_atgte=created_atgte,
                event_types=event_types,
                poll=poll,
                page_size=page_size,
                next_token=next_token,
            )

        def results_extractor(body: EventsPaginatedList) -> Optional[List[Event]]:
            return body.events

        return PageIterator(api_call, results_extractor)



from typing import Iterable, List, Optional, Union

from benchling_api_client.v2.stable.api.organizations import (
    create_organization_membership,
    delete_organization_membership,
    get_organization,
    get_organization_membership,
    list_organization_memberships,
    list_organizations,
    update_organization_membership,
)
from benchling_api_client.v2.types import Response

from benchling_sdk.errors import raise_for_status
from benchling_sdk.helpers.constants import _translate_to_string_enum
from benchling_sdk.helpers.decorators import api_method
from benchling_sdk.helpers.logging_helpers import log_not_implemented
from benchling_sdk.helpers.pagination_helpers import NextToken, PageIterator
from benchling_sdk.helpers.response_helpers import model_from_detailed
from benchling_sdk.helpers.serialization_helpers import none_as_unset, optional_array_query_param
from benchling_sdk.models import (
    ListOrganizationsSort,
    Membership,
    MembershipCreate,
    MembershipsPaginatedList,
    MembershipUpdate,
    Organization,
    OrganizationsPaginatedList,
)
from benchling_sdk.services.v2.base_service import BaseService


class OrganizationService(BaseService):
    """
    Organizations.

    View organization objects.

    See https://benchling.com/api/reference#/Organizations
    """

    @api_method
    def get_by_id(self, organization_id: str) -> Organization:
        """
        Get an organization by ID.

        Returns an organization by ID if the caller has permission to view. The following roles have view permission:
        * tenant admins
        * members of the organization

        See https://benchling.com/api/reference#/Organizations/getOrganization
        """
        response = get_organization.sync_detailed(client=self.client, organization_id=organization_id)
        return model_from_detailed(response)

    @api_method
    def _organizations_page(
        self,
        *,
        ids: Optional[Iterable[str]] = None,
        name: Optional[str] = None,
        name_includes: Optional[str] = None,
        names_any_of: Optional[Iterable[str]] = None,
        names_any_of_case_sensitive: Optional[Iterable[str]] = None,
        modified_at: Optional[str] = None,
        has_members: Optional[Iterable[str]] = None,
        has_admins: Optional[Iterable[str]] = None,
        sort: Optional[ListOrganizationsSort] = None,
        page_size: Optional[int] = 50,
        next_token: Optional[str] = None,
    ) -> Response[OrganizationsPaginatedList]:
        response = list_organizations.sync_detailed(
            client=self.client,
            ids=none_as_unset(optional_array_query_param(ids)),
            name=none_as_unset(name),
            name_includes=none_as_unset(name_includes),
            namesany_of=none_as_unset(optional_array_query_param(names_any_of)),
            namesany_ofcase_sensitive=none_as_unset(optional_array_query_param(names_any_of_case_sensitive)),
            modified_at=none_as_unset(modified_at),
            has_members=none_as_unset(optional_array_query_param(has_members)),
            has_admins=none_as_unset(optional_array_query_param(has_admins)),
            page_size=none_as_unset(page_size),
            next_token=none_as_unset(next_token),
            sort=none_as_unset(sort),
        )
        raise_for_status(response)
        return response  # type: ignore

    def list(
        self,
        *,
        ids: Optional[Iterable[str]] = None,
        name: Optional[str] = None,
        name_includes: Optional[str] = None,
        names_any_of: Optional[Iterable[str]] = None,
        names_any_of_case_sensitive: Optional[Iterable[str]] = None,
        modified_at: Optional[str] = None,
        has_members: Optional[Iterable[str]] = None,
        has_admins: Optional[Iterable[str]] = None,
        sort: Optional[Union[str, ListOrganizationsSort]] = None,
        page_size: Optional[int] = 50,
        mentioned_in: Optional[List[str]] = None,
    ) -> PageIterator[Organization]:
        """
        List organizations.

        Returns all organizations that the caller has permission to view. The following roles have view permission:
        * tenant admins
        * members of the organization

        See https://benchling.com/api/reference#/Organizations/listOrganizations
        """
        if mentioned_in:
            log_not_implemented("mentioned_in")

        def api_call(next_token: NextToken) -> Response[OrganizationsPaginatedList]:
            return self._organizations_page(
                ids=ids,
                name=name,
                name_includes=name_includes,
                names_any_of=names_any_of,
                names_any_of_case_sensitive=names_any_of_case_sensitive,
                modified_at=modified_at,
                has_members=has_members,
                has_admins=has_admins,
                sort=_translate_to_string_enum(ListOrganizationsSort, sort),
                next_token=next_token,
                page_size=page_size,
            )

        def results_extractor(body: OrganizationsPaginatedList) -> Optional[List[Organization]]:
            return body.organizations

        return PageIterator(api_call, results_extractor)

    @api_method
    def _list_memberships_page(
        self,
        organization_id: str,
        role: Optional[str],
        page_size: Optional[int],
        next_token: NextToken,
    ) -> Response[MembershipsPaginatedList]:
        response = list_organization_memberships.sync_detailed(
            client=self.client,
            organization_id=organization_id,
            page_size=none_as_unset(page_size),
            next_token=next_token,
            role=none_as_unset(role),
        )
        raise_for_status(response)
        return response  # type: ignore

    def list_memberships(
        self,
        organization_id: str,
        role: Optional[str] = None,
        page_size: Optional[int] = None,
    ) -> PageIterator[MembershipsPaginatedList]:
        """
        Return all organization memberships in the given organization.

        See https://benchling.com/api/reference#/Organizations/listOrganizationMemberships
        """

        def api_call(next_token: NextToken) -> Response[MembershipsPaginatedList]:
            return self._list_memberships_page(organization_id, role, page_size, next_token)

        return PageIterator(api_call, lambda body: body.memberships)

    @api_method
    def get_membership(self, organization_id: str, user_id: str) -> Membership:
        """
        Get organization membership.

        See https://benchling.com/api/reference#/Organizations/getOrganizationMembership
        """
        response = get_organization_membership.sync_detailed(
            client=self.client, organization_id=organization_id, user_id=user_id
        )
        return model_from_detailed(response)

    @api_method
    def create_membership(self, organization_id: str, membership: MembershipCreate) -> Membership:
        """
        Create organization membership.

        See https://benchling.com/api/reference#/Organizations/createOrganizationMembership
        """
        response = create_organization_membership.sync_detailed(
            client=self.client,
            organization_id=organization_id,
            json_body=membership,
        )
        return model_from_detailed(response)

    @api_method
    def update_membership(
        self, organization_id: str, user_id: str, membership: MembershipUpdate
    ) -> Membership:
        """
        Update organization membership by user ID and organization ID.

        See https://benchling.com/api/reference#/Organizations/updateOrganizationMembership
        """
        response = update_organization_membership.sync_detailed(
            client=self.client,
            organization_id=organization_id,
            user_id=user_id,
            json_body=membership,
        )
        return model_from_detailed(response)

    @api_method
    def delete_membership(self, organization_id: str, user_id: str) -> None:
        """
        Delete a single team membership.

        See https://benchling.com/api/reference#/Organizations/deleteOrganizationMembership
        """
        raise_for_status(
            delete_organization_membership.sync_detailed(
                client=self.client, organization_id=organization_id, user_id=user_id
            )
        )



from typing import Iterable, List, Union

from benchling_api_client.v2.stable.api.registry import (
    bulk_get_registered_entities,
    get_registry,
    list_registries,
    register_entities,
    unregister_entities,
)
from benchling_api_client.v2.stable.extensions import UnknownType

from benchling_sdk.helpers.decorators import api_method
from benchling_sdk.helpers.logging_helpers import check_for_csv_bug_fix
from benchling_sdk.helpers.response_helpers import model_from_detailed
from benchling_sdk.helpers.serialization_helpers import array_query_param
from benchling_sdk.helpers.task_helpers import EmptyTaskResponse, TaskHelper
from benchling_sdk.models import (
    AaSequenceWithEntityType,
    CustomEntityWithEntityType,
    DnaOligoWithEntityType,
    DnaSequenceWithEntityType,
    MixtureWithEntityType,
    NamingStrategy,
    RegisterEntities,
    Registry,
    RnaOligoWithEntityType,
    UnregisterEntities,
)
from benchling_sdk.services.v2.base_service import BaseService

TypedEntity = Union[
    CustomEntityWithEntityType,
    DnaSequenceWithEntityType,
    AaSequenceWithEntityType,
    MixtureWithEntityType,
    DnaOligoWithEntityType,
    RnaOligoWithEntityType,
    UnknownType,
]


class RegistryService(BaseService):
    """
    Registry.

    Manage registry objects.

    See https://benchling.com/api/reference#/Registry
    """

    @api_method
    def register(
        self,
        registry_id: str,
        entity_ids: Iterable[str],
        naming_strategy: NamingStrategy = NamingStrategy.NEW_IDS,
    ) -> TaskHelper[EmptyTaskResponse]:
        """
        Register entities.

        Attempts to move entities into the registry. This endpoint will first check that the entities
        are all valid to be moved into the registry, given the namingStrategy. If any entities fail validation,
        no files will be moved and errors describing invalid entities is returned.

        If all entities pass validation, the entities are moved into the registry.

        See https://benchling.com/api/reference#/Registry/registerEntities
        """
        registration_body = RegisterEntities(entity_ids=list(entity_ids), naming_strategy=naming_strategy)
        response = register_entities.sync_detailed(
            client=self.client, registry_id=registry_id, json_body=registration_body
        )
        return self._task_helper_from_response(response, EmptyTaskResponse)

    @api_method
    def unregister(
        self,
        registry_id: str,
        entity_ids: Iterable[str],
        folder_id: str,
    ) -> None:
        """
        Unregister entities.

        Unregisters entities and moves them to a folder.

        See https://benchling.com/api/reference#/Registry/unregisterEntities
        """
        registration_body = UnregisterEntities(entity_ids=list(entity_ids), folder_id=folder_id)
        response = unregister_entities.sync_detailed(
            client=self.client, registry_id=registry_id, json_body=registration_body
        )
        # Raise for error but return nothing
        model_from_detailed(response)

    @api_method
    def registries(self) -> List[Registry]:
        """
        List registries.

        See https://benchling.com/api/reference#/Registry/listRegistries

        :return: A list of registries
        :rtype: List[Registry]
        """
        response = list_registries.sync_detailed(client=self.client)
        registry_list = model_from_detailed(response)
        if not registry_list.registries:
            return []
        return registry_list.registries

    # TODO Currently this payload does not deserialize properly until
    # https://github.com/triaxtec/openapi-python-client/issues/219

    @api_method
    def entities(self, registry_id: str, entity_registry_ids: Iterable[str]) -> List[TypedEntity]:
        """
        Bulk get registered entities.

        See https://benchling.com/api/reference#/Registry/bulkGetRegisteredEntities
        """
        check_for_csv_bug_fix("entity_registry_ids", entity_registry_ids)

        response = bulk_get_registered_entities.sync_detailed(
            client=self.client,
            registry_id=registry_id,
            entity_registry_ids=array_query_param(entity_registry_ids),
        )
        entity_list = model_from_detailed(response)
        if not entity_list.entities:
            return []
        return entity_list.entities

    @api_method
    def get_by_id(self, registry_id: str) -> Registry:
        """
        Get registry.

        See https://benchling.com/api/reference#/Registry/getRegistry
        """
        response = get_registry.sync_detailed(client=self.client, registry_id=registry_id)
        return model_from_detailed(response)



from __future__ import annotations

from functools import cached_property
from typing import Optional, TYPE_CHECKING

from benchling_api_client.v2.stable.client import Client

from benchling_sdk.helpers.client_helpers import v2_alpha_client
from benchling_sdk.helpers.retry_helpers import RetryStrategy

if TYPE_CHECKING:
    from benchling_sdk.services.v2.alpha.v2_alpha_app_service import V2AlphaAppService
    from benchling_sdk.services.v2.alpha.v2_alpha_assembly_service import V2AlphaAssemblyService

from benchling_sdk.services.v2.base_service import BaseService


class V2AlphaService(BaseService):
    """
    V2-alpha.

    Alpha endpoints have different stability guidelines than other stable endpoints.

    See https://benchling.com/api/v2-alpha/reference
    """

    _alpha_client: Client

    def __init__(self, client: Client, retry_strategy: Optional[RetryStrategy] = None):
        """
        Initialize a v2-alpha service.

        :param client: Underlying generated Client.
        :param retry_strategy: Retry strategy for failed HTTP calls
        """
        super().__init__(client, retry_strategy)
        self._alpha_client = v2_alpha_client(self.client)

    @cached_property
    def apps(self) -> V2AlphaAppService:
        """
        V2-Alpha Apps.

        Create and manage Apps on your tenant.

        https://benchling.com/api/v2-alpha/reference?stability=not-available#/Apps
        """
        from .alpha.v2_alpha_app_service import V2AlphaAppService

        return self._create_service(V2AlphaAppService)

    @cached_property
    def assemblies(self) -> V2AlphaAssemblyService:
        """
        V2-Alpha Assemblies.

        In Benchling, Assemblies are records of a process in which many fragment sequences are
        assembled in silico to create new construct sequences.

            https://benchling.com/api/v2-alpha/reference#/Assemblies
        """
        from .alpha.v2_alpha_assembly_service import V2AlphaAssemblyService

        return self._create_service(V2AlphaAssemblyService)

    def _create_service(self, cls):
        """Instantiate a service using the alpha client."""
        return cls(self._alpha_client, self._retry_strategy)



from typing import Iterable, List, Optional

from benchling_api_client.v2.stable.api.workflow_task_groups import (
    archive_workflow_task_groups,
    create_workflow_task_group,
    get_workflow_task_group,
    list_workflow_task_groups,
    unarchive_workflow_task_groups,
    update_workflow_task_group,
)
from benchling_api_client.v2.types import Response

from benchling_sdk.errors import raise_for_status
from benchling_sdk.helpers.decorators import api_method
from benchling_sdk.helpers.pagination_helpers import NextToken, PageIterator
from benchling_sdk.helpers.response_helpers import model_from_detailed
from benchling_sdk.helpers.serialization_helpers import none_as_unset, optional_array_query_param
from benchling_sdk.models import (
    WorkflowTaskGroup,
    WorkflowTaskGroupArchiveReason,
    WorkflowTaskGroupCreate,
    WorkflowTaskGroupsArchivalChange,
    WorkflowTaskGroupsArchive,
    WorkflowTaskGroupsPaginatedList,
    WorkflowTaskGroupsUnarchive,
    WorkflowTaskGroupUpdate,
)
from benchling_sdk.services.v2.base_service import BaseService


class WorkflowTaskGroupService(BaseService):
    """
    Workflow Task Groups.

    Workflow task groups are groups of workflow tasks of the same schema.

    See https://benchling.com/api/reference#/Workflow%20Task%20Groups
    """

    @api_method
    def get_by_id(self, workflow_task_group_id: str) -> WorkflowTaskGroup:
        """
        Get a workflow task group.

        See https://benchling.com/api/reference#/Workflow%20Task%20Groups/getWorkflowTaskGroup
        """
        response = get_workflow_task_group.sync_detailed(
            client=self.client, workflow_task_group_id=workflow_task_group_id
        )
        return model_from_detailed(response)

    @api_method
    def _workflow_task_groups_page(
        self,
        ids: Optional[Iterable[str]] = None,
        schema_id: Optional[str] = None,
        folder_id: Optional[str] = None,
        project_id: Optional[str] = None,
        mentioned_in: Optional[Iterable[str]] = None,
        watcher_ids: Optional[Iterable[str]] = None,
        execution_types: Optional[Iterable[str]] = None,
        responsible_team_ids: Optional[Iterable[str]] = None,
        status_ids_any_of: Optional[Iterable[str]] = None,
        status_ids_none_of: Optional[Iterable[str]] = None,
        status_ids_only: Optional[Iterable[str]] = None,
        name: Optional[str] = None,
        name_includes: Optional[str] = None,
        creator_ids: Optional[Iterable[str]] = None,
        modified_at: Optional[str] = None,
        display_ids: Optional[Iterable[str]] = None,
        archive_reason: Optional[str] = None,
        page_size: Optional[int] = None,
        next_token: Optional[NextToken] = None,
    ) -> Response[WorkflowTaskGroupsPaginatedList]:
        response = list_workflow_task_groups.sync_detailed(
            client=self.client,
            ids=none_as_unset(optional_array_query_param(ids)),
            schema_id=none_as_unset(schema_id),
            folder_id=none_as_unset(folder_id),
            project_id=none_as_unset(project_id),
            mentioned_in=none_as_unset(optional_array_query_param(mentioned_in)),
            watcher_ids=none_as_unset(optional_array_query_param(watcher_ids)),
            execution_types=none_as_unset(optional_array_query_param(execution_types)),
            responsible_team_ids=none_as_unset(optional_array_query_param(responsible_team_ids)),
            status_idsany_of=none_as_unset(optional_array_query_param(status_ids_any_of)),
            status_idsnone_of=none_as_unset(optional_array_query_param(status_ids_none_of)),
            status_idsonly=none_as_unset(optional_array_query_param(status_ids_only)),
            name=none_as_unset(name),
            name_includes=none_as_unset(name_includes),
            creator_ids=none_as_unset(optional_array_query_param(creator_ids)),
            modified_at=none_as_unset(modified_at),
            display_ids=none_as_unset(optional_array_query_param(display_ids)),
            archive_reason=none_as_unset(archive_reason),
            page_size=none_as_unset(page_size),
            next_token=none_as_unset(next_token),
        )
        raise_for_status(response)
        return response  # type: ignore

    def list(
        self,
        ids: Optional[Iterable[str]] = None,
        schema_id: Optional[str] = None,
        folder_id: Optional[str] = None,
        project_id: Optional[str] = None,
        mentioned_in: Optional[Iterable[str]] = None,
        watcher_ids: Optional[Iterable[str]] = None,
        execution_types: Optional[Iterable[str]] = None,
        responsible_team_ids: Optional[Iterable[str]] = None,
        status_ids_any_of: Optional[Iterable[str]] = None,
        status_ids_none_of: Optional[Iterable[str]] = None,
        status_ids_only: Optional[Iterable[str]] = None,
        name: Optional[str] = None,
        name_includes: Optional[str] = None,
        creator_ids: Optional[Iterable[str]] = None,
        modified_at: Optional[str] = None,
        display_ids: Optional[Iterable[str]] = None,
        archive_reason: Optional[str] = None,
        page_size: Optional[int] = None,
    ) -> PageIterator[WorkflowTaskGroup]:
        """
        List workflow task groups.

        See https://benchling.com/api/reference#/Workflow%20Task%20Groups/listWorkflowTaskGroups
        """

        def api_call(next_token: NextToken) -> Response[WorkflowTaskGroupsPaginatedList]:
            return self._workflow_task_groups_page(
                ids=ids,
                schema_id=schema_id,
                folder_id=folder_id,
                project_id=project_id,
                mentioned_in=mentioned_in,
                watcher_ids=watcher_ids,
                execution_types=execution_types,
                responsible_team_ids=responsible_team_ids,
                status_ids_any_of=status_ids_any_of,
                status_ids_none_of=status_ids_none_of,
                status_ids_only=status_ids_only,
                name=name,
                name_includes=name_includes,
                creator_ids=creator_ids,
                modified_at=modified_at,
                display_ids=display_ids,
                archive_reason=archive_reason,
                page_size=page_size,
                next_token=next_token,
            )

        def results_extractor(body: WorkflowTaskGroupsPaginatedList) -> Optional[List[WorkflowTaskGroup]]:
            return body.workflow_task_groups

        return PageIterator(api_call, results_extractor)

    @api_method
    def create(self, workflow_task_group: WorkflowTaskGroupCreate) -> WorkflowTaskGroup:
        """
        Create a new workflow task group.

        If no name is specified, uses the workflow schema name and a unique incrementor
        separated by a single whitespace.

        See https://benchling.com/api/reference#/Workflow%20Task%20Groups/createWorkflowTaskGroup
        """
        response = create_workflow_task_group.sync_detailed(client=self.client, json_body=workflow_task_group)
        return model_from_detailed(response)

    @api_method
    def update(
        self, workflow_task_group_id: str, workflow_task_group: WorkflowTaskGroupUpdate
    ) -> WorkflowTaskGroup:
        """
        Update a workflow task group.

        See https://benchling.com/api/reference#/Workflow%20Task%20Groups/updateWorkflowTaskGroup
        """
        response = update_workflow_task_group.sync_detailed(
            client=self.client, workflow_task_group_id=workflow_task_group_id, json_body=workflow_task_group
        )
        return model_from_detailed(response)

    @api_method
    def archive(
        self, workflow_task_group_ids: Iterable[str], reason: WorkflowTaskGroupArchiveReason
    ) -> WorkflowTaskGroupsArchivalChange:
        """
        Archive one or more workflows.

        See https://benchling.com/api/reference#/Workflow%20Task%20Groups/archiveWorkflowTaskGroups
        """
        archive_request = WorkflowTaskGroupsArchive(
            reason=reason, workflow_task_group_ids=list(workflow_task_group_ids)
        )
        response = archive_workflow_task_groups.sync_detailed(client=self.client, json_body=archive_request)
        return model_from_detailed(response)

    @api_method
    def unarchive(self, workflow_task_group_ids: Iterable[str]) -> WorkflowTaskGroupsArchivalChange:
        """
        Unarchive one or more workflows.

        See https://benchling.com/api/reference#/Workflow%20Task%20Groups/unarchiveWorkflowTaskGroups
        """
        unarchive_request = WorkflowTaskGroupsUnarchive(workflow_task_group_ids=list(workflow_task_group_ids))
        response = unarchive_workflow_task_groups.sync_detailed(
            client=self.client, json_body=unarchive_request
        )
        return model_from_detailed(response)



from benchling_api_client.v2.stable.api.workflow_flowchart_config_versions import (
    get_workflow_flowchart_config_version,
)

from benchling_sdk.helpers.decorators import api_method
from benchling_sdk.helpers.response_helpers import model_from_detailed
from benchling_sdk.models import WorkflowFlowchartConfigVersion
from benchling_sdk.services.v2.base_service import BaseService


class WorkflowFlowchartConfigVersionService(BaseService):
    """
    Workflow Flowchart Config Versions.

    Workflow flowchart config versions are versioned graphs of flowchart configurations.

    See https://benchling.com/api/reference#/Workflow%20Flowchart%20Config%20Versions
    """

    @api_method
    def get_by_id(self, workflow_flowchart_config_version_id: str) -> WorkflowFlowchartConfigVersion:
        """
        Get a workflow flowchart config version.

        If there is a template flowchart, serializes that flowchart in the same format as the workflow_flowcharts service.

        See https://benchling.com/api/reference#/Workflow%20Flowcharts/getWorkflowConfigVersion
        """
        response = get_workflow_flowchart_config_version.sync_detailed(
            client=self.client, workflow_flowchart_config_version_id=workflow_flowchart_config_version_id
        )
        return model_from_detailed(response)



from typing import Any, Dict, Iterable, List, Optional

from benchling_api_client.v2.stable.api.workflow_outputs import (
    archive_workflow_outputs,
    bulk_create_workflow_outputs,
    bulk_update_workflow_outputs,
    create_workflow_output,
    get_workflow_output,
    list_workflow_outputs,
    unarchive_workflow_outputs,
    update_workflow_output,
)
from benchling_api_client.v2.types import Response

from benchling_sdk.errors import raise_for_status
from benchling_sdk.helpers.decorators import api_method
from benchling_sdk.helpers.pagination_helpers import NextToken, PageIterator
from benchling_sdk.helpers.response_helpers import model_from_detailed
from benchling_sdk.helpers.serialization_helpers import (
    none_as_unset,
    optional_array_query_param,
    schema_fields_query_param,
)
from benchling_sdk.models import (
    AsyncTaskLink,
    WorkflowOutput,
    WorkflowOutputArchiveReason,
    WorkflowOutputBulkCreate,
    WorkflowOutputBulkUpdate,
    WorkflowOutputCreate,
    WorkflowOutputsArchivalChange,
    WorkflowOutputsArchive,
    WorkflowOutputsBulkCreateRequest,
    WorkflowOutputsBulkUpdateRequest,
    WorkflowOutputsPaginatedList,
    WorkflowOutputsUnarchive,
    WorkflowOutputUpdate,
)
from benchling_sdk.services.v2.base_service import BaseService


class WorkflowOutputService(BaseService):
    """
    Workflow Outputs.

    Workflow outputs are outputs of a workflow task.

    See https://benchling.com/api/reference#/Workflow%20Outputs
    """

    @api_method
    def get_by_id(self, workflow_output_id: str) -> WorkflowOutput:
        """
        Get a workflow output.

        See https://benchling.com/api/reference#/Workflow%20Outputs/getWorkflowOutput
        """
        response = get_workflow_output.sync_detailed(
            client=self.client, workflow_output_id=workflow_output_id
        )
        return model_from_detailed(response)

    @api_method
    def _workflow_outputs_page(
        self,
        ids: Optional[Iterable[str]] = None,
        workflow_task_group_ids: Optional[Iterable[str]] = None,
        workflow_task_ids: Optional[Iterable[str]] = None,
        schema_id: Optional[str] = None,
        watcher_ids: Optional[Iterable[str]] = None,
        responsible_team_ids: Optional[Iterable[str]] = None,
        creation_origin_ids: Optional[Iterable[str]] = None,
        schema_fields: Optional[Dict[str, Any]] = None,
        name: Optional[str] = None,
        name_includes: Optional[str] = None,
        creator_ids: Optional[Iterable[str]] = None,
        modified_at: Optional[str] = None,
        display_ids: Optional[Iterable[str]] = None,
        linked_item_ids_any_of: Optional[Iterable[str]] = None,
        linked_item_ids_all_of: Optional[Iterable[str]] = None,
        linked_item_ids_none_of: Optional[Iterable[str]] = None,
        archive_reason: Optional[str] = None,
        page_size: Optional[int] = None,
        next_token: Optional[NextToken] = None,
    ) -> Response[WorkflowOutputsPaginatedList]:
        response = list_workflow_outputs.sync_detailed(
            client=self.client,
            ids=none_as_unset(optional_array_query_param(ids)),
            workflow_task_group_ids=none_as_unset(optional_array_query_param(workflow_task_group_ids)),
            workflow_task_ids=none_as_unset(optional_array_query_param(workflow_task_ids)),
            schema_id=none_as_unset(schema_id),
            watcher_ids=none_as_unset(optional_array_query_param(watcher_ids)),
            responsible_team_ids=none_as_unset(optional_array_query_param(responsible_team_ids)),
            creation_origin_ids=none_as_unset(optional_array_query_param(creation_origin_ids)),
            schema_fields=none_as_unset(schema_fields_query_param(schema_fields)),
            name=none_as_unset(name),
            name_includes=none_as_unset(name_includes),
            creator_ids=none_as_unset(optional_array_query_param(creator_ids)),
            modified_at=none_as_unset(modified_at),
            display_ids=none_as_unset(optional_array_query_param(display_ids)),
            linked_item_idsany_of=none_as_unset(optional_array_query_param(linked_item_ids_any_of)),
            linked_item_idsall_of=none_as_unset(optional_array_query_param(linked_item_ids_all_of)),
            linked_item_idsnone_of=none_as_unset(optional_array_query_param(linked_item_ids_none_of)),
            archive_reason=none_as_unset(archive_reason),
            page_size=none_as_unset(page_size),
            next_token=none_as_unset(next_token),
        )
        raise_for_status(response)
        return response  # type: ignore

    def list(
        self,
        ids: Optional[Iterable[str]] = None,
        workflow_task_group_ids: Optional[Iterable[str]] = None,
        workflow_task_ids: Optional[Iterable[str]] = None,
        schema_id: Optional[str] = None,
        watcher_ids: Optional[Iterable[str]] = None,
        responsible_team_ids: Optional[Iterable[str]] = None,
        creation_origin_ids: Optional[Iterable[str]] = None,
        schema_fields: Optional[Dict[str, Any]] = None,
        name: Optional[str] = None,
        name_includes: Optional[str] = None,
        creator_ids: Optional[Iterable[str]] = None,
        modified_at: Optional[str] = None,
        display_ids: Optional[Iterable[str]] = None,
        linked_item_ids_any_of: Optional[Iterable[str]] = None,
        linked_item_ids_all_of: Optional[Iterable[str]] = None,
        linked_item_ids_none_of: Optional[Iterable[str]] = None,
        archive_reason: Optional[str] = None,
        page_size: Optional[int] = None,
    ) -> PageIterator[WorkflowOutput]:
        """
        List workflow outputs.

        See https://benchling.com/api/reference#/Workflow%20Outputs/listWorkflowOutputs
        """

        def api_call(next_token: NextToken) -> Response[WorkflowOutputsPaginatedList]:
            return self._workflow_outputs_page(
                ids=ids,
                workflow_task_group_ids=workflow_task_group_ids,
                workflow_task_ids=workflow_task_ids,
                schema_id=schema_id,
                watcher_ids=watcher_ids,
                responsible_team_ids=responsible_team_ids,
                creation_origin_ids=creation_origin_ids,
                schema_fields=schema_fields,
                name=name,
                name_includes=name_includes,
                creator_ids=creator_ids,
                modified_at=modified_at,
                display_ids=display_ids,
                linked_item_ids_any_of=linked_item_ids_any_of,
                linked_item_ids_all_of=linked_item_ids_all_of,
                linked_item_ids_none_of=linked_item_ids_none_of,
                archive_reason=archive_reason,
                page_size=page_size,
                next_token=next_token,
            )

        def results_extractor(body: WorkflowOutputsPaginatedList) -> Optional[List[WorkflowOutput]]:
            return body.workflow_outputs

        return PageIterator(api_call, results_extractor)

    @api_method
    def create(self, workflow_output: WorkflowOutputCreate) -> WorkflowOutput:
        """
        Create a new workflow output.

        See https://benchling.com/api/reference#/Workflow%20Outputs/createWorkflowOutput
        """
        response = create_workflow_output.sync_detailed(client=self.client, json_body=workflow_output)
        return model_from_detailed(response)

    @api_method
    def update(self, workflow_output_id: str, workflow_output: WorkflowOutputUpdate) -> WorkflowOutput:
        """
        Update a workflow output.

        See https://benchling.com/api/reference#/Workflow%20Outputs/updateWorkflowOutput
        """
        response = update_workflow_output.sync_detailed(
            client=self.client, workflow_output_id=workflow_output_id, json_body=workflow_output
        )
        return model_from_detailed(response)

    @api_method
    def archive(
        self, workflow_output_ids: Iterable[str], reason: WorkflowOutputArchiveReason
    ) -> WorkflowOutputsArchivalChange:
        """
        Archive one or more workflow outputs.

        See https://benchling.com/api/reference#/Workflow%20Outputs/archiveWorkflowOutputs
        """
        archive_request = WorkflowOutputsArchive(reason=reason, workflow_output_ids=list(workflow_output_ids))
        response = archive_workflow_outputs.sync_detailed(client=self.client, json_body=archive_request)
        return model_from_detailed(response)

    @api_method
    def unarchive(self, workflow_output_ids: Iterable[str]) -> WorkflowOutputsArchivalChange:
        """
        Unarchive one or more workflow outputs.

        See https://benchling.com/api/reference#/Workflow%20Outputs/unarchiveWorkflowOutputs
        """
        unarchive_request = WorkflowOutputsUnarchive(workflow_output_ids=list(workflow_output_ids))
        response = unarchive_workflow_outputs.sync_detailed(client=self.client, json_body=unarchive_request)
        return model_from_detailed(response)

    @api_method
    def bulk_create(self, workflow_outputs: Iterable[WorkflowOutputBulkCreate]) -> AsyncTaskLink:
        """
        Create one or more workflow outputs.

        See https://benchling.com/api/reference#/Workflow%20Outputs/bulkCreateWorkflowOutputs
        """
        body = WorkflowOutputsBulkCreateRequest(list(workflow_outputs))
        response = bulk_create_workflow_outputs.sync_detailed(client=self.client, json_body=body)
        return model_from_detailed(response)

    @api_method
    def bulk_update(self, workflow_outputs: Iterable[WorkflowOutputBulkUpdate]) -> AsyncTaskLink:
        """
        Update one or more workflow outputs.

        See https://benchling.com/api/reference#/Workflow%20Outputs/bulkUpdateWorkflowOutputs
        """
        body = WorkflowOutputsBulkUpdateRequest(list(workflow_outputs))
        response = bulk_update_workflow_outputs.sync_detailed(client=self.client, json_body=body)
        return model_from_detailed(response)



class DuplicateBlockIdError(Exception):
    """Error indicating that duplicate ids were present on blocks within a Canvas."""

    pass


class NoMatchingBlocksError(Exception):
    """
    Error indicating that blocks were expected, but none matched.

    Used to prevent requiring developers to handle Optional[_UiBlock] for type safety.
    """

    pass


class MissingResourceIdError(Exception):
    """Error indicating that a resource ID was expected to be present for the operation."""

    pass
